#include "ATG_tcp_server.h"

double LatShift::latShift_0 = 0.0;
double LatShift::latShift_future = 0.0;

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>
#include <iostream>
using namespace std;

using boost::asio::ip::tcp;

const int buffer_size = 2048;
const int length = 2;

boost::mutex mutex;
typedef boost::shared_ptr<tcp::socket> socket_ptr;
std::vector<socket_ptr> sock;

// extern double latShift_0;
// extern double latShift_dt;

/*------------------------------------------------------------------------
 *  ATG TCP server session
 */
void session()
{
  double message[length] = {0.0, 0.0};
  char data[buffer_size];

  while(true)
    {
      boost::this_thread::sleep(boost::posix_time::milliseconds(1));      

      // get back the lateral shift values
      message[0] = LatShift::latShift_0;   // lateral shift at the current time
      message[1] = LatShift::latShift_future;  // lateral shift at the current time + 0.4

//       cout << "lateral shift 0 : " << LatShift::latShift_0 << "  future : " << LatShift::latShift_future << endl;

      // write the state vector into the socket buffer
      mutex.lock();
      for(std::vector<socket_ptr>::iterator it = sock.begin(); it != sock.end(); ++it)
	{
	  try {
	    boost::asio::write(*(*it), boost::asio::buffer(message, length*sizeof(double)));
	  }
	  catch(std::exception e)
	    {
	      std::cout << "client disconnected" << std::endl;
	      sock.erase(it);
	      break;				
	    }
	}
      mutex.unlock();
    }
}


/*------------------------------------------------------------------------
 *  ATG TCP connection server
 */
void server(boost::asio::io_service& io_service, short port)
{
  boost::thread t(session);

  tcp::acceptor a(io_service, tcp::endpoint(tcp::v4(), port));
  for (;;)
  {
    socket_ptr mysock(new tcp::socket(io_service));
    a.accept(*mysock);

    cout << "client connected" << endl;
	mutex.lock();
	sock.push_back(mysock);
	mutex.unlock();
  }
}


/*------------------------------------------------------------------------
 *  Init the TCP Communication with robufast
 */
void initATG_TCPserver(int port)
{
  try
    {
      boost::asio::io_service io_service;
      server(io_service,  port);
    }
  catch (std::exception& e)
    {
      std::cerr << "Exception: " << e.what() << "\n";
    }
}

/*------------------------------------------------------------------------
 *  Stop the TCP Communication with robufast
 */
void stopATG_TCPserver()
{
}
