/**
 ** MPC_avoidObsCntrlTaskCodels.cc
 **
 ** Codels used by the control task MPC_avoidObsCntrlTask
 **
 ** Author: Vivien Delsart
 ** Date: May 25 2011
 **
 **/

#include <portLib.h>

#include "server/MPC_avoidObsHeader.h"

#include <MPC/trajFollow.hpp>
#include <MPC/avoidObs.hpp>
using namespace mpc;

#include <iostream>
#include <cmath>
using namespace std;

/*------------------------------------------------------------------------
 * SetGoToOptionsCntrl  -  control codel of CONTROL request SetGoToOptions
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
SetGoToOptionsCntrl(OA_GOTO_INIT_I *OA_GoTo_Init_I, int *report)
{
  SDI_F->maxDistTgtReached = OA_GoTo_Init_I->maxDistTgtReached;
  SDI_F->maxDistWpReached = OA_GoTo_Init_I->maxDistWpReached;
  SDI_F->maxAngleWpReached = OA_GoTo_Init_I->maxAngleWpReached / 180.0 * M_PI;

  SDI_F->GoTo_fixFinalTH = OA_GoTo_Init_I->fixFinalTH;

  return OK;
}


/*------------------------------------------------------------------------
 * AddWayPointCntrl  -  control codel of CONTROL request AddWayPoint
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
AddWayPointCntrl(WP2D *AddWayPoint_I, int *report)
{
  int i = SDI_F->listWP.nbWP;
  SDI_F->listWP.tab[i].x = AddWayPoint_I->x;
  SDI_F->listWP.tab[i].y = AddWayPoint_I->y;
  SDI_F->listWP.nbWP++;

  return OK;
}


/*------------------------------------------------------------------------
 * StopCntrl  -  control codel of CONTROL request Stop
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
StopCntrl(int *report)
{
  SDI_F->listWP.nbWP = 0;
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  cerr << "mpcAvoidObs : Stop request received" << endl;

  return OK;
}

/*------------------------------------------------------------------------
 * SetATFParams_Ctrl  -  control codel of CONTROL request SetATFParams
 *
 * Description:    
 * Report: OK
 *                 S_MPC_avoidObs_UNVALID_PARAMETERS
 *
 * Returns:    OK or ERROR
 */

STATUS
SetATFParams_Ctrl(ATFPARAMS_I *ATFParams_I, int *report)
{
  if((ATFParams_I->version == 1 || ATFParams_I->version == 2) && 
     ATFParams_I->ds_maxdist_points > 0.0 &&
     ATFParams_I->min_nbitv_trajgen > 0 &&
     ATFParams_I->coef_time >= 1)
    {
      AvoidTrajGen::version = ATFParams_I->version;

      AvoidTrajGen::setDSMaxDistPoints( ATFParams_I->ds_maxdist_points );
      AvoidTrajGen::setMinNbItvTrajgen( ATFParams_I->min_nbitv_trajgen );
      AvoidTrajGen::setCoefTime( ATFParams_I->coef_time );
      return OK;
    }
  else 
    {
      if(ATFParams_I->version != 1 && ATFParams_I->version != 2) { cerr << "AvoidTrajGen::version must be 1 or 2" << endl; }
      if(ATFParams_I->ds_maxdist_points <= 0.0) { cerr << "AvoidTrajGen::ds_maxdist_points must be > 0.0" << endl; }
      if(ATFParams_I->min_nbitv_trajgen <= 0) { cerr << "AvoidTrajGen::min_nbitv_trajgen must be > 0" << endl; }
      if(ATFParams_I->coef_time < 1) { cerr << "AvoidTrajGen::coef_time must be >= 1" << endl; }

      *report = S_MPC_avoidObs_UNVALID_PARAMETERS;
      return ERROR;
    }
}


/*------------------------------------------------------------------------
 * SetRobufastComParams_Ctrl  -  control codel of CONTROL request SetRobufastComParams
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
SetRobufastComParams_Ctrl(ROBUFAST_COM_PARAMS_I *RobufastComParams_I, int *report)
{
  cout << "Setting Robufast communication parameters..." << endl;
  SDI_F->atg_server_port = RobufastComParams_I->port;
  SDI_F->dt_latShift = RobufastComParams_I->dt_latShift;

  *report = OK;
  return OK;
}

/*------------------------------------------------------------------------
 * SetVelFeedbackMode_Ctrl  -  control codel of CONTROL request SetVelFeedbackMode
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
SetVelFeedbackMode_Ctrl(char Vel_fb_mode_I[256], int *report)
{
  cout << "SetVelFeedbackMode_Ctrl 0" << endl;

  if(strcmp(Vel_fb_mode_I, "gps") == 0)
    {
      cout << "SetVelFeedbackMode_Ctrl 1" << endl;

      if(SDI_F->gps_poster_found)
	{
	  SDI_F->vel_fb_mode = VEL_FB_GPS;
	  cout << "Velocity feedback mode : GPS" << endl;
	}
      else
	{
	  SDI_F->vel_fb_mode = VEL_FB_SENTCTRL;
	  cout << "Velocity feedback mode : GPS poster has not been found. Default used : Sent control" << endl;
	}
    }
  else if(strcmp(Vel_fb_mode_I, "sent_control") == 0)
    {
      cout << "SetVelFeedbackMode_Ctrl 2" << endl;

      if(SDI_F->ctrl_poster_found)
	{
	  SDI_F->vel_fb_mode = VEL_FB_SENTCTRL;
	  cout << "Velocity feedback mode : Sent control" << endl;
	}
      else
	{
	  SDI_F->vel_fb_mode = VEL_FB_SENTCTRL;
	  cout << "Velocity feedback mode : Control poster has not been found. Default used : Sent control" << endl;
	}
    }
  else
    {
      SDI_F->vel_fb_mode = VEL_FB_SENTCTRL;
      cout << "Velocity feedback mode : Unknown mode. Default used : Sent control" << endl;
    }
  return OK;
}

// /*------------------------------------------------------------------------
//  * FollowTraj_Ctrl  -  control codel of EXEC request FollowTraj
//  *
//  * Description:    
//  * Report: OK
//  *                 S_MPC_avoidObs_POM_NOT_FOUND
//  *                 S_MPC_avoidObs_STRAJ_NOT_FOUND
//  *                 S_MPC_avoidObs_SYS_UNDEFINED
//  *
//  * Returns:    OK or ERROR
//  */

// STATUS
// FollowTraj_Ctrl(FOLLOWTRAJ_I *FollowTraj_I, int *report)
// {
//   // check if the robotic system has been defined
//   if(sys_type == MPC_UNDEFINED_SYS)
//     {
//       cout << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : Undefined robotic system." << endl;      
//       *report = S_MPC_avoidObs_SYS_UNDEFINED;
//       return ERROR;
//     }
//   else cout << "MPC_avoidObs (TF) : Robotic system found" << endl;

//   // check if the POM poster has been found
//   if(pomPosterId == NULL)
//     {
//       cout << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : POM poster has not been found." << endl;      
//       *report = S_MPC_avoidObs_POM_NOT_FOUND;
//       return ERROR;
//     }
//   else cout << "MPC_avoidObs (TF) : POM found" << endl;

//   // check if the followed trajectory poster has been found (STRAJ)
//   if(!SDI_F->straj_poster_found)
//     {
//       cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : STRAJ poster has not been found 1." << endl;      
//       *report = S_MPC_avoidObs_STRAJ_NOT_FOUND;
//       return ERROR;
//     }

//   // check if the followed trajectory poster has been found (STRAJ)
//   if(strajPosterId == NULL)
//     {
//       cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : STRAJ poster has not been found 2." << endl;      
//       *report = S_MPC_avoidObs_STRAJ_NOT_FOUND;
//       return ERROR;
//     }

//   // check if the DTM poster has been found (if needed)
//   if(FollowTraj_I->stop_before_col && distmapPosterId == NULL)
//     {
//       cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : DISTMAP poster has not been found." << endl;      
//       *report = S_MPC_avoidObs_DISTMAP_NOT_FOUND;
//       return ERROR;
//     }

//   return OK;
// }

/*------------------------------------------------------------------------
 * SetTFParams_Ctrl  -  control codel of CONTROL request SetTFParams
 *
 * Description:    
 * Report: OK
 *                 S_MPC_avoidObs_UNVALID_PARAMETERS
 *
 * Returns:    OK or ERROR
 */

STATUS
SetTFParams_Ctrl(TFPARAMS_I *TFParams_I, int *report)
{
  if(TFParams_I->minDistToObstacles >= 0 && TFParams_I->maxDistToObstacles >= TFParams_I->minDistToObstacles && TFParams_I->minCoefSpeed >= 0 && TFParams_I->minCoefSpeed <= 1)
    {
      TrajFollow::minDistToObstacles_ = TFParams_I->minDistToObstacles;
      TrajFollow::maxDistToObstacles_ = TFParams_I->maxDistToObstacles;
      TrajFollow::minCoefSpeed_ = TFParams_I->minCoefSpeed;
      
      return OK;
    }
  else 
    {
      if(TFParams_I->minDistToObstacles < 0) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->minDistToObstacles must be >= 0" << endl;
      if(TFParams_I->maxDistToObstacles < TFParams_I->minDistToObstacles) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->maxDistToObstacles must be >= TFParams_I->minDistToObstacles" << endl;
      if(TFParams_I->minCoefSpeed < 0 || TFParams_I->minCoefSpeed > 1) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->minCoefSpeed must belong to [0;1]" << endl;
      
      *report = S_MPC_avoidObs_UNVALID_PARAMETERS;
      return ERROR;
    }
}

/*------------------------------------------------------------------------
 * setDistMapOptions_Ctrl  -  control codel of CONTROL request SetDistMapOptions
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
setDistMapOptions_Ctrl(DISTMAPOPTIONS_I *DistMapOptions_I, int *report)
{
  SDI_F->SlopeMapDist = DistMapOptions_I->SlopeMapDist;
  SDI_F->MaxSlope = DistMapOptions_I->MaxSlope;
  SDI_F->MaxVarSlope = DistMapOptions_I->MaxVarSlope;
  SDI_F->SafeDist = DistMapOptions_I->SafeDist;
  
  return OK;
}
