/**
 ** MPC_avoidObsExecTaskCodels.cc
 **
 ** Codels called by execution task MPC_avoidObsExecTask
 **
 ** Author: Vivien Delsart
 ** Date: 
 **
 **/

#include <portLib.h>

#include "server/MPC_avoidObsHeader.h"

#include <MPC/common.hpp>
#include <MPC/system.hpp>
#include <MPC/trajSel.hpp>
#include <MPC/controller.hpp>
#include <MPC/trajFollow.hpp>
#include <MPC/avoidObs.hpp>
#include <MPC/network.hpp>

using namespace mpc;

#include <iostream>
#include <cstring>
#include <vector>
using namespace std;

#include <boost/bind.hpp>
#include <boost/smart_ptr.hpp>
#include <boost/asio.hpp>
#include <boost/thread.hpp>

// #include "ATG_tcp_server.h"
#include "Posters_ttt.h"

#include <GPSNovatel/msgStruct.h>
#include <GPSNovatel/logDataStruct.h>

#define DIST_TARGET_REACHED 0.25

#define TF_DFLT_TIME_HORIZON 2.0

#define DEBUG 0
#define DEBUG_GEN_TRAJ 0

// trajectory selection options
#define MANATRAJSEL_SPG_ACC 9
#define MANATRAJSEL_SPG_AACC 9

#define DFLT_MANATRAJSEL_SPG_ACC2 3
#define DFLT_MANATRAJSEL_SPG_AACC2 7
#define DFLT_MANATRAJSEL_SPG_JERK2 3
#define DFLT_MANATRAJSEL_SPG_AJERK2 5

#define DFLT_TS_WG 1.0                       // Goal position heuritic weight
#define DFLT_TS_WO 0.2                       //  Obstacles heuritic weight
#define DFLT_TS_WH 1.0                       // Goal heading heuritic weight

#define DFLT_TS_SLOPEMAP_DIST 0.3            // Maximal distance considered between two cells of the slope map to compute the slope */
#define DFLT_TS_MAX_SLOPE 0.3*M_PI           // Maximal slope that the robot can climb */
#define DFLT_TS_MAX_VAR_SLOPE 2.0*M_PI       // Maximal slope derivative that the robot can climb */
#define DFLT_TS_SAFE_DIST 0.2                // Distance that the robot must keep between its center of rotation and the obstacle to be safe */
#define DFLT_TS_TIME_TRAJ 7.0                // Duration of the candidate trajectories */
#define DFLT_TS_TIME_TRAJ_OBS_FREE 5.0       // Time of the candidate trajectories until which collision check is performed */

#define DFLT_TS_SAMPLING_CAND_TRAJ 15        // Sampling of the candidate trajectories */
#define DFLT_TS_DIST_OBS_MAX 2.0             // Maximal distance from obstacles until which a repulsive potential is applied */

ROBOTIC_SYS_T sys_type;              // Robotic system type ( dala, dala_simu   ->  MPC_TF_MANA_SYS
                                     //                       otherwise         ->  MPC_TF_UNDEFINED_SYS )

POSTER_ID pomPosterId;               // POM poster ID (containing the localization of the robot)
POSTER_ID ctrlPosterId;              // Control poster ID (containing the current control applied on the robot)
POSTER_ID demPosterId;               // DEM poster ID (containing the local Data Elevation Map of the environment surrounding the robot)
POSTER_ID distmapPosterId;           // Dist Map poster ID
POSTER_ID wpPosterId;                // Waypoint list poster ID
POSTER_ID pathPosterId;              // Path poster id
POSTER_ID gpsPosterId;               // GPS poster id
POSTER_ID strajPosterId;             // Discrete state trajectory poster id

System* sys;                         // the robotic system managed by MPC
STraj* straj;                        // the trajectory followed

SimpleController* controller;        // the controller managed by MPC
TrajFollow* trajFollow;              // the trajectory following process
TrajSel* trajSel;                    // the obstacle avoidance process (by trajectory selection)
AvoidTrajGen* avoidTG;               // the obstacle avoidance process (by trajectory selection)

DEM* localDEM;                       // the local DEM of the robot
DistObsMap* distMap;                 // the local Dist Map of the robot

XYThS_Path* path;                    // path followed by the AvoidTrajGen process

#if DEBUG_GEN_TRAJ
vector<MANA_PSTRAJ> vec_pstraj;      // vector containing the generated trajectories during the AvoidTrajGen process
#endif

// --------------- user posters

static TAB_CAND_PATH* tabCandPath_user;   // table containing the candidate trajectories computed by the trajectory selection module
static DIST_OBS_MAP* distObsMap_user;     // distance to obstacles map poster
static PL_PATH_XY* atf_waypoints_user;    // Waypoints computed by the avoid traj following process


// boost::thread* thread_tcp_server; 
TCP_server* tcp_server;

// double distTgtReached;               // dist up to which the robot is considered to have reached the goal

double ManaCoefAcc;
double ManaCoefAAcc;
double ManaCoefJerk;
double ManaCoefAJerk;

int ManaTrajSel_spg_acc_used;
int ManaTrajSel_spg_aacc_used;
int ManaTrajSel_spg_jerk_used;
int ManaTrajSel_spg_ajerk_used;

// std::list<GENPOS_CART_CONFIG> waypoints;

// --------------- TCP Communication with Robufast

bool atg_tcp_required;
// int atg_server_port;
// extern double latShift_0;
// extern double latShift_dt;

Timer timer_monitor;                 // timer used to monitor the real execution time of the module


/*------------------------------------------------------------------------
 *  Compute the lateral shifts
 */
void fillLateralShift(const ManaState& s0, const ManaPCTraj* pctraj, struct timeval dateBegTraj)
{
  // check if the reference path has been defined
  if(path == NULL) 
    {
      cerr << "fillLateralShift : No reference path has been defined" << endl;
      return;
    }

  // check if the tcp server is running
  if(tcp_server == NULL)
    {
      cerr << "fillLateralShift : No tcp server has been defined" << endl;
      return;
    }

  // compute the current state trajectory to follow
  ManaPSTraj straj(s0, *pctraj);

  // get back the current state and a state in a close future
  struct timeval date;
  Timer::absoluteTime(date);
  double curElapsedTime = Timer::computeElapsedTime(dateBegTraj, date);  

  DD2StateT current_state = straj.at(curElapsedTime);
  DD2StateT future_state = straj.at(curElapsedTime + SDI_F->dt_latShift);

  // compute the lateras shifts as the minimal distance from each state to the current path  
  double msg[2];
  msg[0] = path->minDistFrom(current_state.x(), current_state.y());
  msg[1] = path->minDistFrom(future_state.x(), future_state.y());
//   msg[0] = 0.0;
//   msg[1] = 0.0;
  tcp_server->setMessage(msg, 2);
}

/*------------------------------------------------------------------------
 *  Stop condition of the obstacle avoidance process
 */
bool goalReached(const GENPOS_CART_CONFIG& curCfg, const GENPOS_CART_CONFIG& goalCfg)
{
  // compute the distance between the current and goal configurations
  float dx = curCfg.x - goalCfg.x;
  float dy = curCfg.y - goalCfg.y;
  float sqDist = dx*dx + dy*dy;

  float goalToCur_head = atan2(dy,dx);
  double diffAngle = fabs(computeMinDiffAngle(goalToCur_head, goalCfg.theta));

  if(sqDist < SDI_F->maxDistTgtReached * SDI_F->maxDistTgtReached)
    {
      return true;
    }
  else if(sqDist < SDI_F->maxDistWpReached * SDI_F->maxDistWpReached && diffAngle < SDI_F->maxAngleWpReached)
    {
      return true;
    }
  else return false;
}


/*------------------------------------------------------------------------
 *  poster reading / writing tools
 */

/*
 *  \brief update the localization of the Mana robot
 *  \param the resulting state
 *  \return 0 if the process succeed, -1 else
 */
int updateManaLoc(ManaState &sres)
{
  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_avoidObsPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {  
      return -1;
    }

  // build the current state
  sres.x() = pos.mainToOrigin.euler.x;
  sres.y() = pos.mainToOrigin.euler.y;
  sres.th() = pos.mainToOrigin.euler.yaw;
  sres.v() = 0.0;
  sres.w() = 0.0;

  if(SDI_F->vel_fb_mode == VEL_FB_SENTCTRL)
    {
//       cout << "MPC_avoidObs : vel_fb_mode = VEL_FB_SENTCTRL" << endl;
      // read the current sent control
      GENPOS_CART_SPEED ctrl;
      if(MPC_avoidObsGENPOS_CART_SPEEDPosterRead (ctrlPosterId, &ctrl) == ERROR)
	{
	  return -1;
	}  
      sres.v() = ctrl.v;
      sres.w() = ctrl.w;
    }
  else if(SDI_F->vel_fb_mode == VEL_FB_GPS)
    {
//       cout << "MPC_avoidObs : vel_fb_mode = GPS" << endl;
      OEM4BESTVELb_LOG gps_poster;
      int offset = sizeof(OEM4BESTUTMb_LOG);
      int resRead = posterRead (gpsPosterId, offset, &gps_poster, sizeof(OEM4BESTVELb_LOG));

      sres.v() = gps_poster.horizontalSpeed;
      sres.w() = 0.0;
    }
#if DEBUG
  cout << "AO loc : (" << sres.x() << "," << sres.y() << "," << sres.th() << ")  speed : (" << sres.v() << "," << sres.w() << ")" << endl;
#endif

  return 0;
}


/*------------------------------------------------------------------------
 *  Trajectory following process tools
 */

/*
 *  \brief build a Mana trajectory from the input poster
 *  \return the Mana state trajectory if the process succeed, NULL else
 */
ManaSTraj* buildManaSTraj()
{
  PL_MANA_STRAJ PL_straj;
  if(MPC_avoidObsPL_MANA_STRAJPosterRead (strajPosterId, &PL_straj) == ERROR)
    {
      return NULL;
    }

  ManaSTraj* newSTraj = new ManaSTraj;
  int N = PL_straj.nbPos;
//   cout << "PL_straj.nbPos : " << PL_straj.nbPos << endl;
  for(int i=0; i<N; i++)
    {
      ManaStateT curST(PL_straj.tab[i].x, 
		       PL_straj.tab[i].y, 
		       PL_straj.tab[i].th, 
		       PL_straj.tab[i].v, 
		       PL_straj.tab[i].w, 
		       PL_straj.tab[i].t);

      newSTraj->append(curST);
    }
  return newSTraj;

//   return NULL;
}


/*------------------------------------------------------------------------
 *  User posters functions
 */

/*
 * \brief fill posters displaying the candidate trajectory and the best one found (user update version)
 */
void fillCandTrajPosters_user(ManaTrajSel* trajSel)
{
  const std::vector<PSTraj*>* vecPSTraj = trajSel->getVecPSTraj();
  int M = vecPSTraj->size();

  for(int i=0; i<M; i++)
    {
      // sample the state trajectory
      ManaSTraj* straj = ((ManaPSTraj*)(vecPSTraj->at(i)))->sample(CAND_PATH_MAXSIZE);

      int N = straj->size();
      for(int j=0; j<N; j++)
	{
	  tabCandPath_user->tab[i].tab[j].x = straj->at(j).x();
	  tabCandPath_user->tab[i].tab[j].y = straj->at(j).y();
	  tabCandPath_user->tab[i].tab[j].theta = straj->at(j).th();
	}
      tabCandPath_user->tab[i].nbPos = straj->size();

      delete straj;
    }
  tabCandPath_user->nbPath = vecPSTraj->size();
}

/*
 *  \brief Fill the distance to obstacles map poster from a mpc distmap
 *  \param distMap : the input mpc distmap
 */
int fillDistMapPoster_user(DistObsMap* distMap)
{
  if(distMap == NULL) return -1;

  distObsMap_user->dimX = distMap->nbCols();
  distObsMap_user->dimY = distMap->nbLines();

  distObsMap_user->xOrigin = distMap->xOrigin();
  distObsMap_user->yOrigin = distMap->yOrigin();
  distObsMap_user->xScale = distMap->xScale();
  distObsMap_user->yScale = distMap->yScale();

  for(int i=0; i < distObsMap_user->dimX; i++)
    {
      for(int j=0; j < distObsMap_user->dimY; j++)
	{
	  obs2D_cell_t cellSt;
	  int distX;
	  int distY;
	  int nearX;
	  int nearY;
	  distMap->getCellInfo(i, j, cellSt, distX, distY, nearX, nearY);

 	  distObsMap_user->cellState[i][j] = cellSt;

	  distObsMap_user->dist[i][j].x = distX;
	  distObsMap_user->dist[i][j].y = distY;

// 	  distObsMap_user->near[i][j].x = nearX;
// 	  distObsMap_user->near[i][j].y = nearY;
	}
    }

  distObsMap_user->thresSlope = distMap->thresSlope();
  distObsMap_user->thresVarSlope = distMap->thresVarSlope();
  distObsMap_user->safeDist = distMap->safeDist();

 //   distMap->saveIntoPPM("DistMap_in_poster.ppm", -15, 22);
  return 0;
}

/*
 *  \brief fill the poster containing the waypoints computed by the mpc avoid traj following process
 *  \param mpc_waypoints : mpc waypoints
 */ 
void fillATFwaypoints_user(const std::vector<WayPoint>& mpc_waypoints)
{
  int N = mpc_waypoints.size();

  for(int i=0; i<N; i++)
    {
      XYThS_Config cfg = mpc_waypoints.at(i).cfg;

      atf_waypoints_user->tab[i].x = cfg.x();
      atf_waypoints_user->tab[i].y = cfg.y();
      atf_waypoints_user->tab[i].theta = cfg.th();
    }

  atf_waypoints_user->nbPos = N;
}

/*------------------------------------------------------------------------*/


/*
 *  Init the trajectory selection process options
 */
void initDfltTrajSelOptions()
{
  TrajSel::TS_SLOPEMAP_DIST = DFLT_TS_SLOPEMAP_DIST;
  TrajSel::TS_MAX_SLOPE = DFLT_TS_MAX_SLOPE;
  TrajSel::TS_MAX_VAR_SLOPE = DFLT_TS_MAX_VAR_SLOPE;
  TrajSel::TS_SAFE_DIST = DFLT_TS_SAFE_DIST;

  TrajSel::TS_TIME_TRAJ = DFLT_TS_TIME_TRAJ;
  TrajSel::TS_TIME_TRAJ_OBS_FREE = DFLT_TS_TIME_TRAJ_OBS_FREE;

  TrajSel::TS_SAMPLING_CAND_TRAJ = DFLT_TS_SAMPLING_CAND_TRAJ;
  TrajSel::TS_DIST_OBS_MAX = DFLT_TS_DIST_OBS_MAX;
  TrajSel::TS_DIST_OBS_DECREASED_SPD = DFLT_TS_DIST_OBS_MAX;

  ManaTrajSel::TS_WG = DFLT_TS_WG;
  ManaTrajSel::TS_WO = DFLT_TS_WO;
  ManaTrajSel::TS_WH = DFLT_TS_WH;

  ManaCoefAcc = 0.3;
  ManaCoefAAcc = 0.5;
  ManaCoefJerk = 2.8;
  ManaCoefAJerk = 4.2;

  ManaTrajSel_spg_acc_used = DFLT_MANATRAJSEL_SPG_ACC2;
  ManaTrajSel_spg_aacc_used = DFLT_MANATRAJSEL_SPG_AACC2;
  ManaTrajSel_spg_jerk_used = DFLT_MANATRAJSEL_SPG_JERK2;
  ManaTrajSel_spg_ajerk_used = DFLT_MANATRAJSEL_SPG_AJERK2;
}

void fillMANA_PCTRAJ_Poster(const ManaPCTraj* pctraj, struct timeval date)
{
  if(pctraj != NULL)
    {
//       cout << "||||||||||||||||||||||||||||||||||||||||||||| AVOIDOBS TRAJ : |||||||||||||||||||||||||||||||||||||||||||||" << endl;
//       pctraj->display();
//       cout << "|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||" << endl;

      vector<vector<double> > coefPolyVel;
      vector<vector<double> > coefPolyAVel;
      double binf;
      double bsup;

//       pctraj->getPolyCoef(coefPolyVel, coefPolyAVel, binf, bsup);
      pctraj->getPolyCoef(coefPolyVel, coefPolyAVel);

      double dt = pctraj->endTime() - pctraj->startTime();
      SDI_F->ctraj.t0.sec = date.tv_sec;
      SDI_F->ctraj.t0.usec = date.tv_usec;
      SDI_F->ctraj.tf.sec = date.tv_sec + ((int)dt);
      SDI_F->ctraj.tf.usec = date.tv_usec + (dt - ((int)dt))*1000000.0;

//       cout << "fillMANA_PCTRAJ_Poster -> dt : " << dt << "   SDI_F->ctraj.t0.sec : " << SDI_F->ctraj.t0.sec << "   SDI_F->ctraj.t0.usec : " << SDI_F->ctraj.t0.usec 
// 	   << "   SDI_F->ctraj.tf.sec : " << SDI_F->ctraj.tf.sec << "   SDI_F->ctraj.tf.usec : " << SDI_F->ctraj.tf.usec << endl;
      
      int VEL_N = coefPolyVel.size();
      if(VEL_N > MAX_NB_ITV) VEL_N = MAX_NB_ITV;

      for(int i=0; i<VEL_N; i++)
	{
	  int curSize = coefPolyVel.at(i).size();

	  SDI_F->ctraj.vel.tabItv[i].itv.binf = coefPolyVel.at(i).at(0);
	  SDI_F->ctraj.vel.tabItv[i].itv.bsup = coefPolyVel.at(i).at(1);

	  if(curSize >= 3) SDI_F->ctraj.vel.tabItv[i].poly.a = coefPolyVel.at(i).at(2);
	  else SDI_F->ctraj.vel.tabItv[i].poly.a = 0;
	  if(curSize >= 4) SDI_F->ctraj.vel.tabItv[i].poly.b = coefPolyVel.at(i).at(3);
	  else SDI_F->ctraj.vel.tabItv[i].poly.b = 0;
	  if(curSize >= 5) SDI_F->ctraj.vel.tabItv[i].poly.c = coefPolyVel.at(i).at(4);
	  else SDI_F->ctraj.vel.tabItv[i].poly.c = 0;
	  if(curSize >= 6) SDI_F->ctraj.vel.tabItv[i].poly.d = coefPolyVel.at(i).at(5);
	  else SDI_F->ctraj.vel.tabItv[i].poly.d = 0;
	}
      SDI_F->ctraj.vel.n = VEL_N;

//       cout << "fillMANA_PCTRAJ_Poster 11111 -> dt : " << dt << "   SDI_F->ctraj.t0.sec : " << SDI_F->ctraj.t0.sec << "   SDI_F->ctraj.t0.usec : " << SDI_F->ctraj.t0.usec 
// 	   << "   SDI_F->ctraj.tf.sec : " << SDI_F->ctraj.tf.sec << "   SDI_F->ctraj.tf.usec : " << SDI_F->ctraj.tf.usec << endl;

      int AVEL_N = coefPolyAVel.size();
      if(AVEL_N > MAX_NB_ITV) AVEL_N = MAX_NB_ITV;

      for(int i=0; i<AVEL_N; i++)
	{
	  int curSize = coefPolyAVel.at(i).size();

	  SDI_F->ctraj.avel.tabItv[i].itv.binf = coefPolyAVel.at(i).at(0);
	  SDI_F->ctraj.avel.tabItv[i].itv.bsup = coefPolyAVel.at(i).at(1);

	  if(curSize >= 3) SDI_F->ctraj.avel.tabItv[i].poly.a = coefPolyAVel.at(i).at(2);
	  else SDI_F->ctraj.avel.tabItv[i].poly.a = 0;
	  if(curSize >= 4) SDI_F->ctraj.avel.tabItv[i].poly.b = coefPolyAVel.at(i).at(3);
	  else SDI_F->ctraj.avel.tabItv[i].poly.b = 0;
	  if(curSize >= 5) SDI_F->ctraj.avel.tabItv[i].poly.c = coefPolyAVel.at(i).at(4);
	  else SDI_F->ctraj.avel.tabItv[i].poly.c = 0;
	  if(curSize >= 6) SDI_F->ctraj.avel.tabItv[i].poly.d = coefPolyAVel.at(i).at(5);
	  else SDI_F->ctraj.avel.tabItv[i].poly.d = 0;
	}
      SDI_F->ctraj.avel.n = AVEL_N;

//       cout << "fillMANA_PCTRAJ_Poster 22222 -> dt : " << dt << "   SDI_F->ctraj.t0.sec : " << SDI_F->ctraj.t0.sec << "   SDI_F->ctraj.t0.usec : " << SDI_F->ctraj.t0.usec 
// 	   << "   SDI_F->ctraj.tf.sec : " << SDI_F->ctraj.tf.sec << "   SDI_F->ctraj.tf.usec : " << SDI_F->ctraj.tf.usec << "   VEL_N : " << VEL_N << "   AVEL_N : " << AVEL_N << endl;
    }
  else
    {
      SDI_F->ctraj.vel.n = 0;
      SDI_F->ctraj.avel.n = 0;

//       cout << "fillMANA_PCTRAJ_Poster : pctraj is null... " << endl;
    }
}

void fillMANA_PSTRAJ_Poster(const ManaState& s0, const ManaPCTraj* pctraj, struct timeval date)
{
  SDI_F->straj.p0.x = s0.x();
  SDI_F->straj.p0.y = s0.y();
  SDI_F->straj.p0.theta = s0.th();

  if(pctraj != NULL)
    {
      vector<vector<double> > coefPolyVel;
      vector<vector<double> > coefPolyAVel;
      double binf; 
      double bsup;
	
//       pctraj->getPolyCoef(coefPolyVel, coefPolyAVel, binf, bsup);
      pctraj->getPolyCoef(coefPolyVel, coefPolyAVel);

//       SDI_F->straj.pctraj.t0 = pctraj->startTime();
//       SDI_F->straj.pctraj.tf = pctraj->endTime();
      double dt = pctraj->endTime() - pctraj->startTime();
      SDI_F->straj.pctraj.t0.sec = date.tv_sec;
      SDI_F->straj.pctraj.t0.usec = date.tv_usec;
//       SDI_F->straj.pctraj.tf.sec = date.tv_sec + ((int)dt);
//       SDI_F->straj.pctraj.tf.usec = date.tv_usec + (dt - ((int)dt))*1000000.0;

      double t2 = date.tv_sec + 0.000001*date.tv_usec + dt;
      SDI_F->straj.pctraj.tf.sec = (int)t2;
      SDI_F->straj.pctraj.tf.usec = (t2 - ((int)t2))*1000000.0;

      int VEL_N = coefPolyVel.size();
      if(VEL_N > MAX_NB_ITV)
	{
	  cerr << "fillMANA_PSTRAJ_Poster : The number of intervals of the linear velocity profile is too great" << endl;
	  return;
	}

      for(int i=0; i<VEL_N; i++)
	{
// 	  cout << "i : " << i << endl;
// 	  cout << "coefPolyVel.at(i).size() : " << coefPolyVel.at(i).size() << endl;
// 	  cout << "coefPolyVel.at(i).at(0) : " << coefPolyVel.at(i).at(0) << endl;
// 	  cout << "coefPolyVel.at(i).at(1) : " << coefPolyVel.at(i).at(1) << endl;
// 	  cout << "coefPolyVel.at(i).at(2) : " << coefPolyVel.at(i).at(2) << endl;

	  int curSize = coefPolyVel.at(i).size();

	  SDI_F->straj.pctraj.vel.tabItv[i].itv.binf = coefPolyVel.at(i).at(0);
	  SDI_F->straj.pctraj.vel.tabItv[i].itv.bsup = coefPolyVel.at(i).at(1);

	  if(curSize >= 3) SDI_F->straj.pctraj.vel.tabItv[i].poly.a = coefPolyVel.at(i).at(2);
	  else SDI_F->straj.pctraj.vel.tabItv[i].poly.a = 0;
	  if(curSize >= 4) SDI_F->straj.pctraj.vel.tabItv[i].poly.b = coefPolyVel.at(i).at(3);
	  else SDI_F->straj.pctraj.vel.tabItv[i].poly.b = 0;
	  if(curSize >= 5) SDI_F->straj.pctraj.vel.tabItv[i].poly.c = coefPolyVel.at(i).at(4);
	  else SDI_F->straj.pctraj.vel.tabItv[i].poly.c = 0;
	  if(curSize >= 6) SDI_F->straj.pctraj.vel.tabItv[i].poly.d = coefPolyVel.at(i).at(5);
	  else SDI_F->straj.pctraj.vel.tabItv[i].poly.d = 0;
	}
      SDI_F->straj.pctraj.vel.n = VEL_N;

      int AVEL_N = coefPolyAVel.size();
      if(VEL_N > MAX_NB_ITV)
	{
	  cerr << "fillMANA_PSTRAJ_Poster : The number of intervals of the angular velocity profile is too great" << endl;
	  return;
	}

      for(int i=0; i<AVEL_N; i++)
	{
	  int curSize = coefPolyAVel.at(i).size();

	  SDI_F->straj.pctraj.avel.tabItv[i].itv.binf = coefPolyAVel.at(i).at(0);
	  SDI_F->straj.pctraj.avel.tabItv[i].itv.bsup = coefPolyAVel.at(i).at(1);

	  if(curSize >= 3) SDI_F->straj.pctraj.avel.tabItv[i].poly.a = coefPolyAVel.at(i).at(2);
	  else SDI_F->straj.pctraj.avel.tabItv[i].poly.a = 0;
	  if(curSize >= 4) SDI_F->straj.pctraj.avel.tabItv[i].poly.b = coefPolyAVel.at(i).at(3);
	  else SDI_F->straj.pctraj.avel.tabItv[i].poly.b = 0;
	  if(curSize >= 5) SDI_F->straj.pctraj.avel.tabItv[i].poly.c = coefPolyAVel.at(i).at(4);
	  else SDI_F->straj.pctraj.avel.tabItv[i].poly.c = 0;
	  if(curSize >= 6) SDI_F->straj.pctraj.avel.tabItv[i].poly.d = coefPolyAVel.at(i).at(5);
	  else SDI_F->straj.pctraj.avel.tabItv[i].poly.d = 0;
	}
      SDI_F->straj.pctraj.avel.n = AVEL_N;
    }
  else
    {
      SDI_F->straj.pctraj.vel.n = 0;
      SDI_F->straj.pctraj.avel.n = 0;
    }
}

void displayPCTraj()
{
  cout << "AO PCTraj -> vel : ( ";
  int VEL_N = SDI_F->ctraj.vel.n;
  for(int i=0; i<VEL_N; i++)
    {
      cout << "[" << SDI_F->ctraj.vel.tabItv[i].itv.binf << ";" << SDI_F->ctraj.vel.tabItv[i].itv.bsup << "]:";
      cout << "(" << SDI_F->ctraj.vel.tabItv[i].poly.a << "," << SDI_F->ctraj.vel.tabItv[i].poly.b << "," << SDI_F->ctraj.vel.tabItv[i].poly.c << "," << SDI_F->ctraj.vel.tabItv[i].poly.d << ") ";
    }

  cout << ")  avel : ( ";
  int AVEL_N = SDI_F->ctraj.avel.n;
  for(int i=0; i<AVEL_N; i++)
    {
      cout << "[" << SDI_F->ctraj.avel.tabItv[i].itv.binf << ";" << SDI_F->ctraj.avel.tabItv[i].itv.bsup << "]:";
      cout << "(" << SDI_F->ctraj.avel.tabItv[i].poly.a << "," << SDI_F->ctraj.avel.tabItv[i].poly.b << "," << SDI_F->ctraj.avel.tabItv[i].poly.c << "," << SDI_F->ctraj.avel.tabItv[i].poly.d << ") ";
    }

  cout << ")" << endl;
}


/*------------------------------------------------------------------------
 *  waypoints list handling
 */

/* Add a new waypoint in the list */
void addWP(float x, float y)
{
  int i = SDI_F->listWP.nbWP;
  SDI_F->listWP.tab[i].x = x;
  SDI_F->listWP.tab[i].y = y;
  SDI_F->listWP.nbWP++;
}

/* get back the first waypoint of the list. return 0 if the process succeed or -1 if the list is empty */
int getFirstWP(double &x, double &y)
{
  if(SDI_F->listWP.nbWP == 0) return -1;
  else
    {
      x = SDI_F->listWP.tab[0].x;
      y = SDI_F->listWP.tab[0].y;
      return 0;
    }
}

/* remove the first waypoint of the list */
void removeFirstWP()
{
  if(SDI_F->listWP.nbWP == 0) return;  

  for(int i=1; i<=SDI_F->listWP.nbWP-1; i++)
    {
      SDI_F->listWP.tab[i-1].x = SDI_F->listWP.tab[i].x;
      SDI_F->listWP.tab[i-1].y = SDI_F->listWP.tab[i].y;
    }
  SDI_F->listWP.nbWP--;
}

/* clear the waypoint list */
void clearListWP()
{
  SDI_F->listWP.nbWP = 0;
}

/* return the size of the list of waypoints */
int getSizeListWP()
{
  return SDI_F->listWP.nbWP;
}


/*------------------------------------------------------------------------
 *
 * MPC_OA_ExecTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_OA_ExecTaskInit(int *report)
{
  sys_type = MPC_UNDEFINED_SYS;

  pomPosterId = NULL;
  ctrlPosterId = NULL;
  demPosterId = NULL;
  distmapPosterId = NULL;
  wpPosterId = NULL;
  pathPosterId = NULL;
  gpsPosterId = NULL;
  strajPosterId = NULL;

  sys = NULL;
  straj = NULL;

  controller = NULL;
  trajFollow = NULL;
  trajSel = NULL;
  avoidTG = NULL;

  path = NULL;

//   thread_tcp_server = NULL;
  tcp_server = NULL;

  localDEM = NULL;
  distMap = NULL;
//   distTgtReached = 0.1;

  clearListWP();

//   LatShift::latShift_0 = 0.0;
//   LatShift::latShift_future = 0.0;

  return OK;
}


/*------------------------------------------------------------------------
 *
 * MPC_OA_ExecTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_OA_ExecTaskEnd(void)
{
  return OK;
}


/*------------------------------------------------------------------------
 *
 * MPC_OA_ExecTaskPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_OA_ExecTaskPerm(int *report)
{
  return OK;
}


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_POSTER_NOT_FOUND
 */

/* MPC_InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_InitMain(OA_INIT_I *Init_I, int *report)
{
//     bool ctrl_poster_found = false;
//     bool gps_poster_found = false;
//     bool distmap_poster_found = false;

  // check for the localization poster (POM)
  if (posterFind(Init_I->pomPoster, &pomPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find pom poster" << endl;
      *report = S_MPC_avoidObs_POSTER_NOT_FOUND;
      return ETHER;
    }
#if DEBUG
  else cerr << "MPC_avoidObs: pom poster found" << endl;
#endif

  // check for the Data Elevation Map poster (DTM)
  if (posterFind(Init_I->demPoster, &demPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find dem poster" << endl;
      *report = S_MPC_avoidObs_POSTER_NOT_FOUND;
      return ETHER;
    }
#if DEBUG
  else cerr << "MPC_avoidObs: dem poster found" << endl;
#endif

  // check for the Dist Map - optional
  cout << "Init_I->distmapPoster : " << Init_I->distmapPoster << endl;

  if (posterFind(Init_I->distmapPoster, &distmapPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find distmap poster" << endl;
    }
  else 
    {
#if DEBUG
      cerr << "MPC_avoidObs: distmap poster found" << endl;
#endif
      SDI_F->distmap_poster_found = true;
    }

  // check for the Control poster (MPC_controller) - optional
  if (posterFind(Init_I->ctrlPoster, &ctrlPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find control poster" << endl;
    }
  else 
    {
#if DEBUG
      cerr << "MPC_avoidObs: control poster found" << endl;
#endif
      SDI_F->ctrl_poster_found = true;
    }

  // check for the GPS poster (GPSInfo) - optional
  if (posterFind(Init_I->gpsPoster, &gpsPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find gps poster" << endl;
    }
  else 
    {
#if DEBUG
      cerr << "MPC_avoidObs: gps poster found" << endl;
#endif
      SDI_F->gps_poster_found = true;
    }

  // check for the state trajectory poster (MPC_planner ManaSTraj - Not mandatory)
  if (posterFind(Init_I->strajPoster, &strajPosterId) == ERROR) 
    {
      cerr << "MPC_avoidObs: cannot find straj poster" << endl;
    }
  else 
    {
#if DEBUG
      cerr << "MPC_avoidObs: straj poster found" << endl;
#endif
      SDI_F->straj_poster_found = true;
    }

  /* Robotic system type */
  SDI_F->system_type = MPC_UNDEFINED_SYS;
  
  /* Current goal location */
  SDI_F->goal_pos.x = 0;
  SDI_F->goal_pos.y = 0;
  SDI_F->goal_pos.theta = 0;
  
  SDI_F->ctraj.vel.n = 0;
  SDI_F->ctraj.avel.n = 0;
  SDI_F->ctraj.t0.sec = 0;
  SDI_F->ctraj.t0.usec = 0;
  SDI_F->ctraj.tf.sec = 0;
  SDI_F->ctraj.tf.usec = 0;

  SDI_F->straj.pctraj.vel.n = 0;
  SDI_F->straj.pctraj.avel.n = 0;
  SDI_F->straj.pctraj.t0.sec = 0;
  SDI_F->straj.pctraj.t0.usec = 0;
  SDI_F->straj.pctraj.tf.sec = 0;
  SDI_F->straj.pctraj.tf.usec = 0;

  /* MPC_controller running mode */
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

//   /* table used to display the candidate trajectories (only containing their paths) */
//   SDI_F->tabCandPath.nbPath = 0;
  
  /* best trajectory found by the trajectory selection module (only containing its path) */
  SDI_F->bestPath.nbPos = 0;

//   /* waypoints computed by the avoid traj following process */
//   SDI_F->atf_waypoints.nbPos = 0;

  /* List of the successive waypoints to reach */
  SDI_F->listWP.nbWP = 0;

//   SDI_F->maxDistTgtReached = DIST_TARGET_REACHED;
  SDI_F->maxDistTgtReached = Init_I->AO_MAX_DIST_TGT_REACHED;
  SDI_F->maxDistWpReached = 0.0;
  SDI_F->maxAngleWpReached = 0.0;

  SDI_F->GoTo_fixFinalTH = true;

  // init the velocity feedback mode (from sent control or GPS)
  SDI_F->vel_fb_mode = VEL_FB_SENTCTRL;

  // init the trajectory selection options
  initDfltTrajSelOptions();

//   SDI_F->distObsMap.dimX = 0;
//   SDI_F->distObsMap.dimY = 0;
//   SDI_F->distObsMap.thresSlope = 0;
//   SDI_F->distObsMap.thresVarSlope = 0;
//   SDI_F->distObsMap.safeDist = 0;

  // ------------------- user poster initialization

  // true if the display posters must be filled
  SDI_F->FILL_DISPLAY_POSTERS = 1;
  SDI_F->display_time = 0;

  // init poster for candidate trajectories of the trajectory selection process
  tabCandPath_user = (TAB_CAND_PATH*)posterAddr (MPC_AVOIDOBS_CANDPATHTS_POSTER_ID);
  memset (tabCandPath_user, 0, sizeof(TAB_CAND_PATH));

  for(int i=0; i<TAB_CAND_PATH_MAXSIZE; i++)
    {
      tabCandPath_user->tab[i].nbPos = 0;
    }
  tabCandPath_user->nbPath = 0;

  // distance to obstacle map
  distObsMap_user = (DIST_OBS_MAP*)posterAddr (MPC_AVOIDOBS_DISTOBSMAP_POSTER_ID);
  memset (distObsMap_user, 0, sizeof(DIST_OBS_MAP));

  distObsMap_user->dimX = 0;
  distObsMap_user->dimY = 0;
  distObsMap_user->thresSlope = 0;
  distObsMap_user->thresVarSlope = 0;
  distObsMap_user->safeDist = 0;

  /* Obstacle avoidance generic options */
  SDI_F->stop_before_col = Init_I->stop_before_col;

  // waypoints computed by the avoid traj following process
  atf_waypoints_user = (PL_PATH_XY*)posterAddr (MPC_AVOIDOBS_ATFWAYPOINTS_POSTER_ID);
  memset (atf_waypoints_user, 0, sizeof(PL_PATH_XY));

  atf_waypoints_user->nbPos = 0;

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetManaSys
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_AO_SetManaSysExec  -  codel EXEC of SetManaSys
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_AO_SetManaSysExec(OA_SETMANASYS_I *SetManaSys_I, int *report)
{
  sys_type = MPC_MANA_SYS;

  // Reset the local system
  if(sys != NULL) delete sys;
  sys = new ManaSys(SetManaSys_I->vmax,
		    SetManaSys_I->accel,
		    SetManaSys_I->wmax,
		    SetManaSys_I->wmax,
		    SetManaSys_I->body_length,
		    SetManaSys_I->body_width,
		    false);

#if DEBUG
  cerr << "MPC_avoidObs: Mana system configured" << endl;

//   cerr << "MPC_avoidObs -> system : " << endl;
//   sys->display();
#endif

  return ETHER;
}


/*------------------------------------------------------------------------
 * SimpleGoTo
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_NO_PATH
 *              S_MPC_avoidObs_BLOCKED
 *              S_MPC_avoidObs_POSTER_READ
 */

/* SimpleGoToStart  -  codel START of SimpleGoTo
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SimpleGoToStart(TRAJSEL_I *Goto_I, int *report)
{
  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cerr << "MPC_avoidObs: Cannot start the simple goto process : Undefined robotic system." << endl;      
      *report = S_MPC_avoidObs_SYS_UNDEFINED;
      return FAIL;
    }

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cerr << "MPC_avoidObs: Cannot start the simple goto process : POM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_POM_POSTER_READ;
      return FAIL;
    }

  // set the goal location
  SDI_F->goal_pos.x = Goto_I->goal.x;
  SDI_F->goal_pos.y = Goto_I->goal.y;

  cout << "======================================== SDI_F->GoTo_fixFinalTH : " << SDI_F->GoTo_fixFinalTH << endl;

  if(SDI_F->GoTo_fixFinalTH == 0)
    {
      ManaSimpleController::fixFinalTH = false;
      SDI_F->goal_pos.theta = Goto_I->goal.theta / 180.0 * M_PI;
      cout << "======================================== GOTO PROCESS START : the final heading is FREE : " << SDI_F->goal_pos.theta << endl;
    }
  else if(SDI_F->GoTo_fixFinalTH == 1)
    {
      ManaSimpleController::fixFinalTH = true;
      SDI_F->goal_pos.theta = Goto_I->goal.theta / 180.0 * M_PI;
      cout << "======================================== GOTO PROCESS START : the final heading is FIXED : " << SDI_F->goal_pos.theta << endl;
    }
  else //  if(SDI_F->GoTo_fixFinalTH == 2)
    {
      ManaSimpleController::fixFinalTH = true;

      ManaState s1; 
      if(updateManaLoc(s1) != 0) 
	{
	  cout << "GOTO PROCESS START : loc update error" << endl;
	  return FAIL;
	}
      double dx = SDI_F->goal_pos.x - s1.x();
      double dy = SDI_F->goal_pos.y - s1.y();
      SDI_F->goal_pos.theta = atan2(dy,dx);
      cout << "======================================== GOTO PROCESS START : the final heading is RECOMPUTED : " << SDI_F->goal_pos.theta << endl;
    }

  // Reset the controller with a simple controller for a Mana robotic system
  if(controller != NULL) delete controller;
#if DEBUG
  controller = new ManaSimpleController((ManaSys*)sys, 1.0, 1);
#else
  controller = new ManaSimpleController((ManaSys*)sys, 1.0);
#endif

  cout << "GOTO PROCESS STARTED" << endl;

  SDI_F->mode = AVOIDOBS_SIMPLE_GOTO;

  *report = OK;
  return EXEC;
}

/* SimpleGoToMain  -  codel EXEC of SimpleGoTo
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SimpleGoToMain(TRAJSEL_I *Goto_I, int *report)
{
  if(SDI_F->mode == AVOIDOBS_SIMPLE_GOTO)
    {
      struct timeval dateLoc;
      Timer::absoluteTime(dateLoc);

      // update the Mana localization
      ManaState s1; 
      if(updateManaLoc(s1) != 0)
	{
	  // send a null control
	  fillMANA_PCTRAJ_Poster(NULL, dateLoc);

	  return FAIL;
	}

      // get back the goal location and  build the goal state
      ManaState s2(SDI_F->goal_pos.x,
		   SDI_F->goal_pos.y,
		   SDI_F->goal_pos.theta,
		   0.0,
		   0.0);

      // check if the goal position has been reached
      GENPOS_CART_CONFIG curCfg;
      curCfg.x = s1.x();
      curCfg.y = s1.y();
      curCfg.theta = s1.th();
      
      if(goalReached(curCfg, SDI_F->goal_pos))
	{

	  // the goal position has been reached
	  SDI_F->mode = AVOIDOBS_NOT_RUNNING;
	  fillMANA_PCTRAJ_Poster(NULL, dateLoc);

	  if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }

// #if DEBUG
	  cout << "THE GOAL HAS BEEN REACHED" << endl;
// #endif
	  return END;
	}

      // compute the control to apply
      ManaPCTraj* ctraj = ((ManaSimpleController*)controller)->computeCTraj(s1, s2);
      fillMANA_PCTRAJ_Poster(ctraj, dateLoc);

      if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(&s1, ctraj); }

//       displayPCTraj();

      *report = OK;
      return EXEC;
    }
  else
    {
      // send a null control

      struct timeval dateLoc;
      Timer::absoluteTime(dateLoc);

      fillMANA_PCTRAJ_Poster(NULL, dateLoc);

      *report = OK;
      return ETHER;
    }
}

/* SimpleGoToEnd  -  codel END of SimpleGoTo
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SimpleGoToEnd(TRAJSEL_I *Goto_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  cout << "MPC_avoidObs : SIMPLE GOTO END" << endl;
  
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetTrajSelOptions
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_AO_SetTrajSelOptions  -  codel EXEC of SetTrajSelOptions
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_AO_SetTrajSelOptions(OA_TRAJSEL_OPTIONS_I *OA_TrajSel_Options_I, int *report)
{
//   ManaTrajSel::TS_WG = OA_TrajSel_Options_I->AO_WG;
//   ManaTrajSel::TS_WO = OA_TrajSel_Options_I->AO_WO;
//   ManaTrajSel::TS_WH = OA_TrajSel_Options_I->AO_WH;

  SDI_F->maxDistTgtReached = OA_TrajSel_Options_I->AO_MAX_DIST_TGT_REACHED;
  SDI_F->maxDistWpReached = OA_TrajSel_Options_I->AO_MAX_DIST_WP_REACHED;
  SDI_F->maxAngleWpReached = OA_TrajSel_Options_I->AO_MAX_ANGLE_WP_REACHED / 180.0 * M_PI;

  TrajSel::TS_SLOPEMAP_DIST = OA_TrajSel_Options_I->AO_SLOPEMAP_DIST;
  TrajSel::TS_MAX_SLOPE = OA_TrajSel_Options_I->AO_MAX_SLOPE;
  TrajSel::TS_MAX_VAR_SLOPE = OA_TrajSel_Options_I->AO_MAX_VAR_SLOPE;
  TrajSel::TS_SAFE_DIST = OA_TrajSel_Options_I->AO_SAFE_DIST;

  TrajSel::TS_TIME_TRAJ = OA_TrajSel_Options_I->AO_TIME_TRAJ;
  TrajSel::TS_TIME_TRAJ_OBS_FREE = OA_TrajSel_Options_I->AO_TIME_TRAJ_OBS_FREE;

  TrajSel::TS_DIST_OBS_MAX = OA_TrajSel_Options_I->AO_DIST_OBS_MAX;
  TrajSel::TS_DIST_OBS_DECREASED_SPD = OA_TrajSel_Options_I->AO_DIST_OBS_MAX;
  TrajSel::TS_SAMPLING_CAND_TRAJ = OA_TrajSel_Options_I->AO_SAMPLING_CAND_TRAJ;

//   if(OA_TrajSel_Options_I->AO_UNKWN_IS_OBS == 1) DistObsMap::unkwnIsObs_ = true;
//   else DistObsMap::unkwnIsObs_ = false;

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetManaTrajSelOptions
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_AO_SetManaTrajSelOptions  -  codel EXEC of SetManaTrajSelOptions
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_AO_SetManaTrajSelOptions(OA_MANATRAJSEL_OPTIONS_I *OA_ManaTrajSel_Options_I, int *report)
{
  ManaTrajSel::TS_WG = OA_ManaTrajSel_Options_I->AO_WG;
  ManaTrajSel::TS_WO = OA_ManaTrajSel_Options_I->AO_WO;
  ManaTrajSel::TS_WH = OA_ManaTrajSel_Options_I->AO_WH;

  ManaCoefAcc = OA_ManaTrajSel_Options_I->COEF_ACC;
  ManaCoefAAcc = OA_ManaTrajSel_Options_I->COEF_AACC;
  ManaCoefJerk = OA_ManaTrajSel_Options_I->COEF_JERK;
  ManaCoefAJerk = OA_ManaTrajSel_Options_I->COEF_AJERK;

  ManaTrajSel_spg_acc_used = OA_ManaTrajSel_Options_I->ManaTrajSel_spg_acc;
  ManaTrajSel_spg_aacc_used = OA_ManaTrajSel_Options_I->ManaTrajSel_spg_aacc;
  ManaTrajSel_spg_jerk_used = OA_ManaTrajSel_Options_I->ManaTrajSel_spg_jerk;
  ManaTrajSel_spg_ajerk_used = OA_ManaTrajSel_Options_I->ManaTrajSel_spg_ajerk;

  return ETHER;
}


/*------------------------------------------------------------------------
 * TrajSelTo
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_NO_TRAJ
 *              S_MPC_avoidObs_BLOCKED
 *              S_MPC_avoidObs_POM_POSTER_READ
 *              S_MPC_avoidObs_DEM_POSTER_READ
 */

/* MPC_TS_Start  -  codel START of TrajSelTo
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TS_Start(TRAJSEL_I *Goto_I, int *report)
{
//   SDI_F->mode = CTRLER_TRAJECTORY_SELECTION;
  SDI_F->mode = AVOIDOBS_TRAJECTORY_SELECTION;

  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cerr << "MPC_avoidObs: Cannot start the TrajSel process : Undefined robotic system." << endl;      
      *report = S_MPC_avoidObs_SYS_UNDEFINED;
      return FAIL;
    }

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cerr << "MPC_avoidObs: Cannot start the TrajSel process : POM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_POM_POSTER_READ;
      return FAIL;
    }

  // check if the DEM poster has been found
  if(demPosterId == NULL)
    {
      cerr << "MPC_avoidObs: Cannot start the TrajSel process : DEM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_DEM_POSTER_READ;
      return FAIL;
    }

//   // check if the Dist Map poster has been found
//   if(distmapPosterId == NULL)
//     {
//       cerr << "MPC_avoidObs: Cannot start the TrajSel process : Dist Map poster has not been found." << endl;      
//       *report = S_MPC_avoidObs_DEM_POSTER_READ;
//       return FAIL;
//     }

  // Reset the controller with a simple controller for a Mana robotic system
  if(trajSel != NULL) delete trajSel;
//   trajSel = new ManaTrajSel((ManaSys*) sys, MANATRAJSEL_SPG_ACC, MANATRAJSEL_SPG_AACC);
  trajSel = new ManaTrajSel((ManaSys*) sys, ManaTrajSel_spg_acc_used, ManaTrajSel_spg_aacc_used, ManaTrajSel_spg_jerk_used, ManaTrajSel_spg_ajerk_used, ManaCoefAcc, ManaCoefAAcc, ManaCoefJerk, ManaCoefAJerk);

  if(SDI_F->display_time)
    {
      timer_monitor.reinit();
      timer_monitor.start();
    }

  cout << "TRAJSEL PROCESS STARTED" << endl;

  // set the goal location
  SDI_F->goal_pos.x = Goto_I->goal.x;
  SDI_F->goal_pos.y = Goto_I->goal.y;
  SDI_F->goal_pos.theta = Goto_I->goal.theta;

  *report = OK;
  return EXEC;
}

int selectNextGoal()
{
  // get back the size of the current list of waypoints
  int sizeWP = getSizeListWP();

  // check if the list of waypoints to reach is empty
  if(sizeWP == 0) return 0;
  else
    {
      int res = getFirstWP(SDI_F->goal_pos.x, SDI_F->goal_pos.y);
      if(res != 0) { cerr << "MPC_avoidObs::TrajSel : Warning. The list of waypoints is empty. This case should not appear here!" << endl; exit(1); }
      SDI_F->goal_pos.theta = 0;

      // remove the reached waypoint from the list
      removeFirstWP();

      return 1;
    }
}

/* MPC_TS_Exec  -  codel EXEC of TrajSelTo
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TS_Exec(TRAJSEL_I *Goto_I, int *report)
{
  if(SDI_F->display_time) { cout << "MPC_avoidObs::TrajSelTo -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl; }

  // update the Mana localization
  ManaState curS; 
  if(updateManaLoc(curS) != 0)
    {
      // send a null control trajectory
      cout << "MPC_avoidObs : MANA LOC UPDATE FAILED" << endl;

//       fillMANA_PSTRAJ_Poster(ManaState(), NULL);
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);

      return FAIL;
    }

  // get back the goal location
  PositionXY targetPos(SDI_F->goal_pos.x,
		       SDI_F->goal_pos.y);
  
  // check if the goal position has been reached
//   double dx = targetPos.x() - curS.x();
//   double dy = targetPos.y() - curS.y();
//   double dist = sqrt(dx*dx + dy*dy);

//   if(dist < SDI_F->maxDistTgtReached)

  GENPOS_CART_CONFIG curCfg;
  curCfg.x = curS.x();
  curCfg.y = curS.y();
  curCfg.theta = curS.th();
  
  if(goalReached(curCfg, SDI_F->goal_pos))

    {
      // send a null control trajectory
      cout << "MPC_avoidObs : GOAL REACHED" << endl;

//       fillMANA_PSTRAJ_Poster(curS, NULL);
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);
      
#if DEBUG
      cout << "THE GOAL HAS BEEN REACHED" << endl;
#endif
      if(selectNextGoal()) return EXEC;
      else return END;
    }

  // update the robot local Dist Map
  DistObsMap* dom = updateDistMap(distmapPosterId); 
  if(dom == NULL)
    {
      // send a null control trajectory
      cout << "MPC_avoidObs : DIST MAP UPDATE FAILED" << endl;

      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);

      return FAIL;
    }
  if(distMap != NULL) delete distMap;
  distMap = dom;

  // compute the control trajectory to apply
  int resUpDM = trajSel->updateDistMap(distMap);
  if(resUpDM != 0)
    {
      cout << "MPC_avoidObs : TRAJSEL FAILED. UNABLE TO UPDATE THE DIST MAP." << endl;
      return FAIL;
    }

//   // update the current dem
//   DEM* dem = updateDEM(demPosterId);

//   // update the obstacle avoidance process dem
//   int resUpDEM = trajSel->updateDEM(dem);
//   if(resUpDEM != 0)
//     {
//       cout << "MPC_avoidObs : TRAJSEL FAILED. UNABLE TO UPDATE THE DEM." << endl;
//       return FAIL;
//     }
//   delete dem;

  int resTrajSel = ((ManaTrajSel*)trajSel)->computeBestTraj(curS, targetPos);

  if(resTrajSel == -1)
    {
      cout << "MPC_avoidObs : TRAJSEL FAILED. THE CURRENT LOCATION OR THE DEM ARE UNDEFINED." << endl;
      return FAIL;
    }
  else if(resTrajSel == -2)
    {
      cout << "MPC_avoidObs : TRAJSEL END. THE TARGET POSITION IS UNREACHABLE." << endl;
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);
      return END;
    }
  else if(resTrajSel == -3)
    {
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);
      if(SDI_F->stop_before_col) 
	{
	  cout << "MPC_avoidObs : TRAJSEL END. THE ROBOT IS IN COLLISION WITH OBSTACLES. STOPPING THE TRAJSEL PROCESS." << endl;
	  return END;
	}
      else 
	{
	  cout << "MPC_avoidObs : TRAJSEL WARNING. THE ROBOT IS IN COLLISION WITH OBSTACLES. WAITING FOR THE PASSAGE TO BE FREED." << endl;
	  return EXEC;
	}
    }

  // get back the best control trajectory found
//   ManaPCTraj* bestCTraj = ((ManaTrajSel*)trajSel)->getBestCTraj();
  ManaPCTraj* bestCTraj;
  struct timeval date;
  ((ManaTrajSel*)trajSel)->getBestCTraj(bestCTraj, date);

  // fill the poster containing the best control trajectory found
  fillMANA_PCTRAJ_Poster(bestCTraj, date);

  if(SDI_F->FILL_DISPLAY_POSTERS)
    {
      // fill posters displaying the candidate trajectory and the best one found
      fillCandTrajPosters_user((ManaTrajSel*)trajSel);
      fillBestTrajPoster(&curS, bestCTraj, NULL);
    }

  if(bestCTraj != NULL) delete bestCTraj;

  *report = OK;

  return EXEC;
}

/* MPC_TS_End  -  codel END of TrajSelTo
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TS_End(TRAJSEL_I *Goto_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  cout << "MPC_avoidObs : TRAJECTORY SELECTION END" << endl;
//   fillBestCTrajPoster(NULL);

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }
  
  return ETHER;
}

/* MPC_TS_Fail  -  codel FAIL of TrajSelTo
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TS_Fail(TRAJSEL_I *Goto_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  cout << "MPC_avoidObs : TRAJECTORY SELECTION FAIL" << endl;
//   fillBestCTrajPoster(NULL);

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}

/* MPC_TS_Inter  -  codel INTER of TrajSelTo
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_TS_Inter(TRAJSEL_I *Goto_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  cout << "MPC_avoidObs : TRAJECTORY SELECTION INTER" << endl;
//   fillBestCTrajPoster(NULL);

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}


/*------------------------------------------------------------------------
 * Track
 *
 * Description: 
 *
 * Reports:      OK
 */

/* TrackStart  -  codel START of Track
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackStart(TRACK_I *Track_I, int *report)
{
  // check if the obstacle avoidance process is already running
  if(SDI_F->mode != AVOIDOBS_NOT_RUNNING)
    {
      cerr << "MPC_avoidObs : Cannot start the waypoint tracking process. The obstacle avoidance process is already running." << endl;
      return ETHER;
    }

  // check if the input poster is available
  if (MPC_avoidObsGENPOS_TRAJ_POINTSPosterFind(Track_I->wpPoster, &wpPosterId) != OK) 
    {
      cerr << "MPC_avoidObs : Cannot start the waypoint tracking process. Unable to find the waypoint list poster" << endl;
      return ETHER;
    }

  SDI_F->mode = AVOIDOBS_WP_POSTER_TRACKING;
  return EXEC;
}

/* TrackMain  -  codel EXEC of Track
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackMain(TRACK_I *Track_I, int *report)
{
  // read the waypoint list poster
  GENPOS_TRAJ_POINTS wpPoster;
  if (MPC_avoidObsGENPOS_TRAJ_POINTSPosterRead(wpPosterId, &wpPoster) != OK) 
    {
      clearListWP();
      SDI_F->mode = AVOIDOBS_NOT_RUNNING;
      return END;
    }

  // update the current list with it
  clearListWP();
//   for(int i=wpPoster.numRef; i<wpPoster.nbPts; i++)
  for(int i=0; i<wpPoster.nbPts; i++)
    {
      addWP(wpPoster.points[i].x, wpPoster.points[i].y);
    }

  return EXEC;
}

/* TrackEnd  -  codel END of Track
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackEnd(TRACK_I *Track_I, int *report)
{
  clearListWP();
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;
  return ETHER;
}

/* TrackInter  -  codel INTER of Track
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
TrackInter(TRACK_I *Track_I, int *report)
{
  clearListWP();
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;
  return ETHER;
}


/*------------------------------------------------------------------------
 * AvoidTF
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_PATH_POSTER_READ
 *              S_MPC_avoidObs_POM_POSTER_READ
 *              S_MPC_avoidObs_DEM_POSTER_READ
 *              S_MPC_avoidObs_SYS_UNDEFINED
 */

void reinit_AvoidTF(ManaSys* sys, XYThS_Path* path, float max_path_length, float width)
{
  // Reset the trajectory following with obstacle avoidance process
  if(avoidTG != NULL) delete avoidTG;
  avoidTG = new DD2AvoidTrajGen(sys, path, max_path_length, width);
}

/* AvoidTF_Start  -  codel START of AvoidTF
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
AvoidTF_Start(AVOIDTF_I *AvoidTF_I, int *report)
{
//   cout << "AvoidTF_Start deb" << endl;

  SDI_F->mode = AVOIDOBS_AVOID_TRAJGEN;

  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : Undefined robotic system." << endl;      
      *report = S_MPC_avoidObs_SYS_UNDEFINED;
      return END;
    }
  else cout << "MPC_avoidTF: Robotic system found" << endl;

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : POM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_POM_POSTER_READ;
      return END;
    }
  else cout << "MPC_avoidTF: POM found" << endl;

  // check if the DEM poster has been found
  if(demPosterId == NULL)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : DEM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_DEM_POSTER_READ;
      return END;
    }
  else cout << "MPC_avoidTF: DEM found" << endl;

  // check for the path poster (MPC_plannerPathRef)
  if (posterFind(AvoidTF_I->pathPoster, &pathPosterId) == ERROR) 
    {
      cout << "MPC_avoidTF: cannot find path poster" << endl;
      *report = S_MPC_avoidObs_PATH_POSTER_READ;
      return END;
    }
#if DEBUG
  else cout << "MPC_avoidTF: path poster found" << endl;
#endif

  // read the path poster
  PL_PATH_XY path_poster;
  if(MPC_avoidObsPL_PATH_XYPosterRead (pathPosterId, &path_poster) == ERROR)
    {
      cout << "MPC_avoidTF: unable to read the path poster" << endl;
      return END;
    }
  else cout << "MPC_avoidTF: path found" << endl;

  // check if the path is not empty
  if(path_poster.nbPos == 0)
    {
      cout << "MPC_avoidTF: FOLLOWED PATH IS EMPTY" << endl;
      return END;
    }

  // build the path to follow
  DPMPath* dpmpath = PL_PATH_XYstruct_to_DPMPath(path_poster);

  if(path != NULL) delete path;
  path = new XYThS_Path(*dpmpath);
  delete dpmpath;

  cout << "WIDTH : " << AvoidTF_I->width << endl;

  // Init the AvoidTrajGen general options
//   AvoidTrajGen::setDSMaxDistPoints( AvoidTF_I->ds_maxdist_points );
//   AvoidTrajGen::setMinNbItvTrajgen( AvoidTF_I->min_nbitv_trajgen );
//   AvoidTrajGen::setCoefTime( AvoidTF_I->coef_time );

  AvoidTrajGen::SLOPEMAP_DIST = SDI_F->SlopeMapDist;
  AvoidTrajGen::MAX_SLOPE = SDI_F->MaxSlope;
  AvoidTrajGen::MAX_VAR_SLOPE = SDI_F->MaxVarSlope;
  AvoidTrajGen::SAFE_DIST = SDI_F->SafeDist;

  if(SDI_F->FILL_DISPLAY_POSTERS) AvoidTrajGen::POSTER_DEBUG_ATG = true;
  else AvoidTrajGen::POSTER_DEBUG_ATG = false;

  // Reset the trajectory following with obstacle avoidance process
  reinit_AvoidTF((ManaSys*) sys, path, AvoidTF_I->max_path_length, AvoidTF_I->width);
  
//   // starts the tcp server if needed
//   if(atg_tcp_required) initATG_TCPserver(SDI_F->atg_server_port);

#if DEBUG_GEN_TRAJ
  vec_pstraj.clear();
#endif

  if(SDI_F->display_time)
    {
      timer_monitor.reinit();
      timer_monitor.start();
    }

  cout << "AVOIDTF PROCESS STARTED" << endl;

  *report = OK;
  return EXEC;
}

/* AvoidTF_Main  -  codel EXEC of AvoidTF
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
AvoidTF_Main(AVOIDTF_I *AvoidTF_I, int *report)
{
  if(SDI_F->display_time) { cout << "MPC_avoidObs::AvoidTF -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl; }

  Timer timer_avoidtf_main;
  timer_avoidtf_main.start();

  // update the Mana localization
  ManaState curS; 
  int resUpdateManaLoc = updateManaLoc(curS);
//   cout << "resUpdateManaLoc : " << resUpdateManaLoc << endl;

//   if(updateManaLoc(curS) != 0)
  if(resUpdateManaLoc != 0)
    {
      // send a null control trajectory
      cout << "MPC_avoidObs : MANA LOC UPDATE FAILED" << endl;
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);
      if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }

      return END;
    }

  if(SDI_F->display_time) { cout << "AvoidTF_Main -> time updateManaLoc : " << timer_avoidtf_main.elapsedTime() << endl; }

  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // update the current dem
  DEM* dem = updateDEM(demPosterId);

  if(SDI_F->display_time) { cout << "AvoidTF_Main -> time updateDEM : " << timer_avoidtf_main.elapsedTime() << endl; }

  // update the obstacle avoidance process dem
  int resUpDEM = avoidTG->updateDEM(dem);
  delete dem;

  if(resUpDEM != 0)
    {
      cout << "MPC_avoidObs::AvoidTF_Main : DEM update failed" << endl;
      return END;
    }

  if(SDI_F->display_time) { cout << "AvoidTF_Main -> time avoidTG updateDEM : " << timer_avoidtf_main.elapsedTime() << endl; }

  // compute the obstacle avoidance trajectory
  int res = ((DD2AvoidTrajGen*)avoidTG)->computeAvoidTraj(curS);
//   cout << "res avoid trajgen : " << res << endl;

  if(SDI_F->display_time) { cout << "AvoidTF_Main -> time computeAvoidTraj : " << timer_avoidtf_main.elapsedTime() << endl; }
      
  switch(res)
    {
    case 1 : // the goal has been reached
      {
#if DEBUG
	cout << "MPC_avoidObs::AvoidTF_Main : THE GOAL HAS BEEN REACHED" << endl;
#endif
	if(AvoidTF_I->loop_on_traj)
	  {
	    // Reset the trajectory following with obstacle avoidance process
	    reinit_AvoidTF((ManaSys*) sys, path, AvoidTF_I->max_path_length, AvoidTF_I->width);
	    
	    return EXEC;
	  }
	else
	  {
	    // send a null control trajectory
	    fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	    if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }
	    
	    return END;
	  }
      }
    case 0 : // standard case
      {
	// get back the control trajectory to apply
	DD2PCTraj* pctraj = (DD2PCTraj*)(avoidTG->getPCTraj());
	if(SDI_F->display_time) { cout << "AvoidTF_Main -> time getPCTraj : " << timer_avoidtf_main.elapsedTime() << endl; }
	
	if(pctraj == NULL) 
	  {
	    cout << "PCTRAJ IS NULL" << endl;
	    return END;
	  }
	fillMANA_PCTRAJ_Poster(pctraj, dateLoc);
	if(SDI_F->display_time) { cout << "AvoidTF_Main -> time fillMANA_PCTRAJ_Poster : " << timer_avoidtf_main.elapsedTime() << endl; }

	DistObsMap* distMap = NULL;
	if(SDI_F->FILL_DISPLAY_POSTERS) 
	  {
	    distMap = avoidTG->distmap();
	    fillBestTrajPoster(&curS, pctraj, distMap); 
	  }
	
	if(SDI_F->display_time) { cout << "AvoidTF_Main -> time fillBestTrajPoster : " << timer_avoidtf_main.elapsedTime() << endl; }
	
#if DEBUG_GEN_TRAJ
	MANA_PSTRAJ mana_pstraj;
	convert_mpcStatePCTraj_into_MANA_PSTRAJ(curS, *pctraj, dateLoc, mana_pstraj);
	
	vec_pstraj.push_back(mana_pstraj);
	cout << "AvoidTF_Main -> time convertPSTraj : " << timer_avoidtf_main.elapsedTime() << endl;
#endif
	
	if(SDI_F->FILL_DISPLAY_POSTERS)
	  {
	    std::vector<WayPoint> mpc_waypoints = avoidTG->waypoints();
	    fillATFwaypoints_user(mpc_waypoints);
	    if(SDI_F->display_time) { cout << "AvoidTF_Main -> time fill waypoints poster : " << timer_avoidtf_main.elapsedTime() << endl; }
	    
// 	    fillDistMapPoster_user(avoidTG->distmap());
	    fillDistMapPoster_user(distMap);
	    if(SDI_F->display_time) { cout << "AvoidTF_Main -> time fill distmap poster : " << timer_avoidtf_main.elapsedTime() << endl; }
	    delete distMap;
	  }
	
	// set the lateral shift if needed
	if(atg_tcp_required) fillLateralShift(curS, pctraj, dateLoc);
	delete pctraj;
	
	if(SDI_F->display_time) { cout << "AvoidTF_Main -> time end : " << timer_avoidtf_main.elapsedTime() << endl; }
	return EXEC;
	break;
      }

    case -2 : 
    case -3 : 
      {
	cerr << "MPC_avoidObs : AvoidTrajGen failed (bad configuration or I/O problem). Stopping the system." << endl;
	// send a null control
	fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }
	// stop the AvoidTF process
	return END;
	break;
      }

    case -1 : 
      {
	cerr << "MPC_avoidObs : AvoidTrajGen : The trajectory generation process failed to find a complete avoidance trajectory." << endl;
	// send a null control
	fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }
	// stop the AvoidTF process
	return END;
	break;	
      }

    case -4 : 
      {
	cerr << "MPC_avoidObs : AvoidTrajGen : A collision may occur. Stopping the system." << endl;
	// send a null control
	fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }
	// stop the AvoidTF process
	return END;
	break;
      }

    default : 
      {
	cerr << "MPC_avoidObs : unknown AvoidTrajGen return status" << endl;
	// send a null control
	fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }
	// stop the AvoidTF process
	return END;
	break;
      }
    }
}

/* AvoidTF_End  -  codel END of AvoidTF
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
AvoidTF_End(AVOIDTF_I *AvoidTF_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // send a null control trajectory
  cout << "MPC_avoidObs : AVOIDTF TRAJGEN END" << endl;
  fillMANA_PCTRAJ_Poster(NULL, dateLoc);
  if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }

#if DEBUG_GEN_TRAJ
  // save the set of computed trajectories into an output file
  saveATF_Trajs_in_file(vec_pstraj, "ATFTrajs.log");

  vec_pstraj.clear();
#endif

//   // stops the tcp server if needed
//   if(atg_tcp_required) stopATG_TCPserver();

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}

/* AvoidTF_Fail  -  codel FAIL of AvoidTF
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
AvoidTF_Fail(AVOIDTF_I *AvoidTF_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // send a null control trajectory
  cout << "MPC_avoidObs : AVOIDOBS TRAJGEN FAIL" << endl;
  fillMANA_PCTRAJ_Poster(NULL, dateLoc);
  if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }

//   // stops the tcp server if needed
//   if(atg_tcp_required) stopATG_TCPserver();

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}

/* AvoidTF_Inter  -  codel INTER of AvoidTF
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
AvoidTF_Inter(AVOIDTF_I *AvoidTF_I, int *report)
{
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // send a null control trajectory
  cout << "MPC_avoidObs : AVOIDOBS TRAJGEN INTER" << endl;
  fillMANA_PCTRAJ_Poster(NULL, dateLoc);
  if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(NULL, NULL); }

//   // stops the tcp server if needed
//   if(atg_tcp_required) stopATG_TCPserver();

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}


/*------------------------------------------------------------------------
 * KeepLatShift
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_PATH_POSTER_READ
 *              S_MPC_avoidObs_POM_POSTER_READ
 *              S_MPC_avoidObs_DEM_POSTER_READ
 *              S_MPC_avoidObs_SYS_UNDEFINED
 */

/* KeepLatShift_Start  -  codel START of KeepLatShift
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
KeepLatShift_Start(KEEP_SHIFT_I *Keep_shift_I, int *report)
// KeepLatShift_Start(int *report)
{
  cout << "KEEP LAT SHIFT START" << endl;

  SDI_F->mode = AVOIDOBS_AVOID_TRAJGEN;

  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : Undefined robotic system." << endl;      
      *report = S_MPC_avoidObs_SYS_UNDEFINED;
      return FAIL;
    }
  else cout << "MPC_avoidTF: Robotic system found" << endl;

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : POM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_POM_POSTER_READ;
      return FAIL;
    }
  else cout << "MPC_avoidTF: POM found" << endl;

  // check if the DEM poster has been found
  if(demPosterId == NULL)
    {
      cout << "MPC_avoidTF: Cannot start the AvoidTF process : DEM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_DEM_POSTER_READ;
      return FAIL;
    }
  else cout << "MPC_avoidTF: DEM found" << endl;

//   // check for the path poster (MPC_plannerPathRef)
//   if (posterFind(AvoidTF_I->pathPoster, &pathPosterId) == ERROR) 
//     {
//       cout << "MPC_avoidTF: cannot find path poster" << endl;
//       *report = S_MPC_avoidObs_PATH_POSTER_READ;
//       return FAIL;
//     }
// #if DEBUG
//   else cout << "MPC_avoidTF: path poster found" << endl;
// #endif

//   // read the path poster
//   PL_PATH_XY path_poster;
//   if(MPC_avoidObsPL_PATH_XYPosterRead (pathPosterId, &path_poster) == ERROR)
//     {
//       cout << "MPC_avoidTF: unable to read the path poster" << endl;
//       return FAIL;
//     }
//   else cout << "MPC_avoidTF: path found" << endl;

  return EXEC;
}

/* KeepLatShift_Main  -  codel EXEC of KeepLatShift
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
KeepLatShift_Main(KEEP_SHIFT_I *Keep_shift_I, int *report)
// KeepLatShift_Main(int *report)
{
  cout << "KEEP LAT SHIFT MAIN" << endl;

  // update the robot localization
  ManaState curS; 
  if(updateManaLoc(curS) != 0)
    {
      // send a null control trajectory
      cout << "MPC_avoidObs : LOC UPDATE FAILED" << endl;
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      fillMANA_PCTRAJ_Poster(NULL, tv);

      return FAIL;
    }

  double msg[2];
  msg[0] = Keep_shift_I->shift;
  msg[1] = Keep_shift_I->shift;
  tcp_server->setMessage(msg, 2);

  if(atg_tcp_required) tcp_server->setMessage(msg, 2);

  return EXEC;
}

/* KeepLatShift_End  -  codel END of KeepLatShift
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
// KeepLatShift_End(int *report)
KeepLatShift_End(KEEP_SHIFT_I *Keep_shift_I, int *report)
{
  cout << "KEEP LAT SHIFT END" << endl;
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  return ETHER;
}

/* KeepLatShift_Fail  -  codel FAIL of KeepLatShift
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
KeepLatShift_Fail(KEEP_SHIFT_I *Keep_shift_I, int *report)
// KeepLatShift_Fail(int *report)
{
  cout << "KEEP LAT SHIFT FAIL" << endl;
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  return ETHER;
}

/* KeepLatShift_Inter  -  codel INTER of KeepLatShift
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
KeepLatShift_Inter(KEEP_SHIFT_I *Keep_shift_I, int *report)
// KeepLatShift_Inter(int *report)
{
  cout << "KEEP LAT SHIFT INTER" << endl;
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  return ETHER;
}


/*------------------------------------------------------------------------
 * StartATG_tcp_server
 *
 * Description: 
 *
 * Reports:      OK
 */

/* StartATG_tcp_server  -  codel EXEC of StartATG_tcp_server
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartATG_tcp_server(int *report)
{
  cout << "Starting ATG tcp server..." << endl;
  atg_tcp_required = true;

  // stop and replace the previous tcp server if needed
  if(tcp_server != NULL)
    {
      tcp_server->stop();
      delete   tcp_server;
    }
  tcp_server = new TCP_server(2, SDI_F->atg_server_port);

  // send an initial null shift
  double msg[2];
  msg[0] = 0.0;
  msg[1] = 0.0;
  tcp_server->setMessage(msg, 2);

  // start the tcp communication
  tcp_server->start();

//   thread_tcp_server = new boost::thread(initATG_TCPserver, SDI_F->atg_server_port);
//   initATG_TCPserver(SDI_F->atg_server_port);
  return ETHER;
}

/*------------------------------------------------------------------------
 * StopATG_tcp_server
 *
 * Description: 
 *
 * Reports:      OK
 */

/* StopATG_tcp_server  -  codel EXEC of StopATG_tcp_server
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StopATG_tcp_server(int *report)
{
  cout << "Stoping ATG tcp server..." << endl;
  atg_tcp_required = false;
//   stopATG_TCPserver();

  if(tcp_server != NULL)
    {
      tcp_server->stop();
      delete   tcp_server;
      tcp_server = NULL;
    }

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetTFParams
 *
 * Description: 
 *
 * Reports:      OK
 */

/* SetTFParams_Ctrl  -  codel EXEC of SetTFParams
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SetTFParams_Main(TFPARAMS_I *TFParams_I, int *report)
{
  if(TFParams_I->minDistToObstacles >= 0 && TFParams_I->maxDistToObstacles >= TFParams_I->minDistToObstacles && TFParams_I->minCoefSpeed >= 0 && TFParams_I->minCoefSpeed <= 1)
    {
      TrajFollow::minDistToObstacles_ = TFParams_I->minDistToObstacles;
      TrajFollow::maxDistToObstacles_ = TFParams_I->maxDistToObstacles;
      TrajFollow::minCoefSpeed_ = TFParams_I->minCoefSpeed;
      
      return ETHER;
    }
  else 
    {
      if(TFParams_I->minDistToObstacles < 0) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->minDistToObstacles must be >= 0" << endl;
      if(TFParams_I->maxDistToObstacles < TFParams_I->minDistToObstacles) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->maxDistToObstacles must be >= TFParams_I->minDistToObstacles" << endl;
      if(TFParams_I->minCoefSpeed < 0 || TFParams_I->minCoefSpeed > 1) cerr << "MPC_avoidObs::SetTFParams_Main : TFParams_I->minCoefSpeed must belong to [0;1]" << endl;
 
     return FAIL;
    }
}

/*------------------------------------------------------------------------
 * FollowTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_avoidObs_POM_NOT_FOUND
 *              S_MPC_avoidObs_STRAJ_NOT_FOUND
 *              S_MPC_avoidObs_SYS_UNDEFINED
 */

void reinit_TrajFollow(ManaSys* syst, ManaSTraj* strajy, int horizon_mode, float horizon)
{
  TrajFollow::maxDistTgtReached_ = SDI_F->maxDistTgtReached;
  TrajFollow::maxDistWpReached_ = SDI_F->maxDistWpReached;
  TrajFollow::maxAngleWpReached_ = SDI_F->maxAngleWpReached;

  // create a trajectory following process to follow the new trajectory
  if(trajFollow != NULL) delete trajFollow;
  trajFollow = new ManaTrajFollow(syst, strajy);
  
  // check if the input time horizon is consistent
  if(horizon_mode == DIST_HORIZON_MODE && horizon > 0)
    {
      // set the time horizon of the trajectory following process
      trajFollow->horizonMode() = TF_DIST_HORIZON_MODE;
      trajFollow->distReach() = horizon;
    }
  else if(horizon_mode == TIME_HORIZON_MODE && horizon > 0)
    {
      // set the time horizon of the trajectory following process
      trajFollow->horizonMode() = TF_TIME_HORIZON_MODE;
      trajFollow->dtReach() = horizon;
    }
  else 
    {
      trajFollow->dtReach() = TF_DFLT_TIME_HORIZON;
      cout << "MPC_avoidObs (TF) : Warning. Input time horizon is not consistent (must be > 0). Default value used : " << TF_DFLT_TIME_HORIZON << endl;
    }
  cout << "TF time horizon : " << trajFollow->dtReach() << endl;
}

/* FollowTraj_Start  -  codel START of FollowTraj
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
FollowTraj_Start(FOLLOWTRAJ_I* FollowTraj_I, int *report)
{
  // check if the robotic system has been defined
  if(sys_type == MPC_UNDEFINED_SYS)
    {
      cout << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : Undefined robotic system." << endl;      
      *report = S_MPC_avoidObs_SYS_UNDEFINED;
      return END;
    }
  else cout << "MPC_avoidObs (TF) : Robotic system found" << endl;

  // check if the POM poster has been found
  if(pomPosterId == NULL)
    {
      cout << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : POM poster has not been found." << endl;      
      *report = S_MPC_avoidObs_POM_NOT_FOUND;
      return END;
    }
  else cout << "MPC_avoidObs (TF) : POM found" << endl;

  // check if the followed trajectory poster has been found (STRAJ)
  if(!SDI_F->straj_poster_found)
    {
      cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : STRAJ poster has not been found 1." << endl;      
      *report = S_MPC_avoidObs_STRAJ_NOT_FOUND;
      return END;
    }

  // check if the followed trajectory poster has been found (STRAJ)
  if(strajPosterId == NULL)
    {
      cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : STRAJ poster has not been found 2." << endl;      
      *report = S_MPC_avoidObs_STRAJ_NOT_FOUND;
      return END;
    }

  // check if the DTM poster has been found (if needed)
  if(FollowTraj_I->stop_before_col && distmapPosterId == NULL)
    {
      cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : DISTMAP poster has not been found." << endl;      
      *report = S_MPC_avoidObs_DISTMAP_NOT_FOUND;
      return END;
    }

  // build a trajectory from the straj poster
  if(straj != NULL) delete straj;
  straj = buildManaSTraj();
//   straj = NULL;
  if(straj == NULL)
    {
      cerr << "MPC_avoidObs (TF) : Cannot start the FollowTraj process : The state trajectory building failed." << endl;      
      return END;
    }
#if DEBUG
  else 
    {
      cout << "Followed state trajectory : " << endl;
      straj->display(); 
      cout << endl;
    }
#endif

#if DEBUG
  cout << "FOLLOWED TRAJECTORY RETRIEVED" << endl;
#endif 

  // remove trajgen debug display
  TrajGen::setDBG(0);

  // reinit the trajectory following process
  reinit_TrajFollow((ManaSys*)sys, (ManaSTraj*)straj, FollowTraj_I->horizon_mode, FollowTraj_I->horizon);

  trajFollow->start();

  // set the current status as started
  SDI_F->mode = AVOIDOBS_SIMPLE_FOLLOW_TRAJ;

  if(SDI_F->display_time)
    {
      timer_monitor.reinit();
      timer_monitor.start();
    }

  cout << "TRAJECTORY FOLLOWING PROCESS STARTED" << endl;

  *report = OK;
  return EXEC;
}

/* FollowTraj_Main  -  codel EXEC of FollowTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
FollowTraj_Main(FOLLOWTRAJ_I* FollowTraj_I, int *report)
{
  if(SDI_F->display_time)
    {
      cout << "MPC_avoidObs::FollowTraj -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl;
    }

  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // update the Mana localization
  ManaState curS; 
  if(updateManaLoc(curS) != 0)
    {
      // send a null control trajectory
      fillMANA_PCTRAJ_Poster(NULL, dateLoc);
      return END;
    }

  if(FollowTraj_I->stop_before_col)
    {
      // update the robot local Dist Map
      DistObsMap* dom = updateDistMap(distmapPosterId); 
      if(dom == NULL)
	{
	  // send a null control trajectory
	  cout << "MPC_avoidObs : DIST MAP UPDATE FAILED" << endl;
	  
	  struct timeval tv;
	  tv.tv_sec = 0;
	  tv.tv_usec = 0;
	  fillMANA_PCTRAJ_Poster(NULL, tv);
	  
	  return FAIL;
	}
      if(distMap != NULL) delete distMap;
      distMap = dom;

      // update the distmap of the trajectory following process
      int resUpDM = trajFollow->updateDistMap(distMap);
      if(resUpDM != 0)
	{
	  cout << "MPC_avoidObs : TRAJFOLLOW FAILED. UNABLE TO UPDATE THE DIST MAP." << endl;
	  return FAIL;
	}
    }

  // get back the control to apply to follow the trajectory
  ManaControl ctrl;
  ManaPCTraj* ctraj = NULL;

  try
    {
//       res = ((ManaTrajFollow*)trajFollow)->getControl(curS, ctrl);
      ctraj = ((ManaTrajFollow*)trajFollow)->getCTraj(curS);
    }
  catch(std::string str)
    {
      cout << "========================================= OA FOLLOWTRAJ CAUGHT EXCEPTION : " << str << endl;
      exit(1);
    }
  
  if(ctraj == NULL)
    {
      // send a null control trajectory
      fillMANA_PCTRAJ_Poster(NULL, dateLoc);
#if DEBUG
      cout << "WARNING : THE TRAJECTORY GENERATION PROCESS FAILED TO JOIN THE FOLLOWED TRAJECTORY!" << endl;
#endif
      return EXEC;
    }

  else if(trajFollow->endTraj())
    {
      // check if the robot must loop on the followed trajectory
      if(FollowTraj_I->loop_on_traj)
	{
	  // send a null control trajectory
	  fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	  
	  // stops the current trajectory following process
	  trajFollow->stop();

	  // reinit a new trajectory following process
	  reinit_TrajFollow((ManaSys*)sys, (ManaSTraj*)straj, FollowTraj_I->horizon_mode, FollowTraj_I->horizon);

	  // starts the new trajectory following process
	  trajFollow->start();

	  return EXEC;
	}
      else
	{
	  // the goal position has been reached
	  SDI_F->mode = AVOIDOBS_NOT_RUNNING;
	  // send a null control trajectory
	  fillMANA_PCTRAJ_Poster(NULL, dateLoc);
	  // #if DEBUG
	  cout << "THE END OF THE TRAJECTORY HAS BEEN REACHED" << endl;
	  // #endif
	  *report = OK;
	  return END;
	}
    }

  else
    {
      // send the control trajectory computed
      fillMANA_PCTRAJ_Poster(ctraj, dateLoc);
      
      if(SDI_F->FILL_DISPLAY_POSTERS) { fillBestTrajPoster(&curS, ctraj, NULL); }

      *report = OK;
      return EXEC;
    }

  return ETHER;
}

/* FollowTraj_End  -  codel END of FollowTraj
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
FollowTraj_End(FOLLOWTRAJ_I* FollowTraj_I, int *report)
{
  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // set the current status as stopped
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  fillMANA_PCTRAJ_Poster(NULL, dateLoc);

  // stops the trajectory following process
  trajFollow->stop();

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ETHER;
}

/* FollowTraj_Fail  -  codel FAIL of FollowTraj
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
FollowTraj_Fail(FOLLOWTRAJ_I* FollowTraj_I, int *report)
{
  // get back the current time
  struct timeval dateLoc;
  Timer::absoluteTime(dateLoc);

  // set the current status as stopped
  SDI_F->mode = AVOIDOBS_NOT_RUNNING;

  // send a null control trajectory
  fillMANA_PCTRAJ_Poster(NULL, dateLoc);

  // stops the trajectory following process
  trajFollow->stop();

  if(SDI_F->display_time)
    {
      timer_monitor.stop();
    }

  return ZOMBIE;
}
