#include "Posters_ttt.h"



// /*------------------------------------------------------------------------
//  *  Compute the lateral shifts
//  */
// void fillLateralShift(const ManaState& s0, const ManaPCTraj* pctraj, struct timeval dateBegTraj)
// {
//   // check if the reference path has been defined
//   if(path == NULL) 
//     {
//       cerr << "fillLateralShift : No reference path has been defined" << endl;
//       return;
//     }

//   // compute the current state trajectory to follow
//   ManaPSTraj straj(s0, *pctraj);

//   // get back the current state and a state in a close future
//   struct timeval date;
//   Timer::absoluteTime(date);
//   double curElapsedTime = Timer::computeElapsedTime(dateBegTraj, date);  

//   DD2StateT current_state = straj.at(curElapsedTime);
//   DD2StateT future_state = straj.at(curElapsedTime + SDI_F->dt_latShift);

//   // compute the lateras shifts as the minimal distance from each state to the current path
//   LatShift::latShift_0 = path->minDistFrom(current_state.x(), current_state.y());
//   LatShift::latShift_future = path->minDistFrom(future_state.x(), future_state.y());
// //   LatShift::latShift_0 = 0.0;
// //   LatShift::latShift_future = 0.0;
// }

/*------------------------------------------------------------------------
 *  poster reading / writing tools
 */

/**
 *  \brief Convert a PL_PATH_XY poster into a DPMPath
 *  \param poster : the input poster
 *  \return a new DPMPath corresponding to the input poster
 *
 *  A new instance of DPMPath object is created
 */
DPMPath* PL_PATH_XYstruct_to_DPMPath(const PL_PATH_XY& poster)
{
  // create a new path
  DPMPath* newPath = new DPMPath();

  int N = poster.nbPos;

  for(int i=0; i<N; i++)
    {
      // get back the current configuration of the path
      GENPOS_CART_CONFIG pos = poster.tab[i];

      // create a MPC configuration from it
      DPMConfig cfg(pos.x,pos.y,pos.theta);

      // append it to the new path
      newPath->append(cfg);
    }

  // return the built path
  return newPath;
}

/**
 *  \brief Convert a DTM_P3D poster into a DEM
 *  \param poster : the input poster
 *  \return a new DEM corresponding to the input poster
 *
 *  A new instance of DEM object is created
 */
DEM* DTM_P3D_POSTER_to_DEM(const DTM_P3D_POSTER& poster)
{
  DEM* dem = new DEM(poster.nbCols,
		     poster.nbLines,
		     poster.xOrigin,
		     poster.yOrigin,
		     poster.zOrigin,
		     poster.xScale,
		     poster.yScale,
		     poster.zScale);

  int NB_COLS = dem->nbCols();
  int NB_LINES = dem->nbLines();

  for(int i=0; i<NB_COLS; i++)
    {
      for(int j=0; j<NB_LINES; j++)
	{
// 	  if(poster.nb_points[i][j] != 0)
// 	    {
// 	      dem->setCell(i, 
// 			   j, 
// 			   poster.zfloat[i][j], 
// 			   poster.nb_points[i][j], 
// 			   poster.sigma_z[i][j], 
// 			   poster.confidence[i][j]);
// 	    }
	  if(poster.state[i][j] == DTM_CELL_FILLED)
	    {
	      dem->setCell(i, 
			   j, 
			   poster.zfloat[i][j], 
			   1, 
			   0, 
			   0);
	    }
	  else // if(poster.state[i][j] != DTM_CELL_EMPTY)
	    {
	      dem->setEmptyCell(i, j);
	    }
	}
    }

  return dem;
}

/*
 *  \brief update the local DEM of the Mana robot
 *  \return the resulting DEM if the process succeed, NULL else
 */
DEM* updateDEM(const POSTER_ID& demPosterId)
{
  // read the current DEM
  DTM_P3D_POSTER dem;
  if(MPC_avoidObsDTM_P3D_POSTERPosterRead (demPosterId, &dem) == ERROR)
    {
      return NULL;
    }

  // convert it into a mpc::DEM class
  DEM* demRes = DTM_P3D_POSTER_to_DEM(dem);
  return demRes;
}

/*
 *  \brief Convert a DIST_OBS_MAP genom structure into a mpc DistObsMap class
 *  \param poster : the input genom DIST_OBS_MAP poster
 *  \return a mpc distance to obstacles map
 */
DistObsMap* DIST_OBS_MAP_POSTER_to_DISTMAP(const DIST_OBS_MAP& poster)
{
  DistObsMap* distmap = new DistObsMap(poster.dimX, poster.dimY, 
				       poster.xOrigin, poster.yOrigin, poster.xScale, poster.yScale, 
				       poster.thresSlope, poster.thresVarSlope, poster.safeDist);

  for(int i=0; i<poster.dimX; i++)
    for(int j=0; j<poster.dimY; j++)
      {
 	obs2D_cell_t state;
	if(poster.cellState[i][j] == CELL_UNKWN) state = CELL_UNKWN;
	else if(poster.cellState[i][j] == CELL_OBS) state = CELL_OBS;
	else if(poster.cellState[i][j] == CELL_FREE) state = CELL_FREE;
	
	pos_cell_xy dist;
	pos_cell_xy near;
	dist.x = poster.dist[i][j].x;
	dist.y = poster.dist[i][j].y;
//  	near.x = poster.near[i][j].x;
//  	near.y = poster.near[i][j].y;
 	near.x = 0;
 	near.y = 0;

	distmap->setDistAndState(i, j, state, dist, near);

// 	if(poster.cellState[i][j] == CELL_UNKWN) distmap->setUndefCell(i,j);
// 	else
// 	  {
// 	    pos_cell_xy dist;
// 	    pos_cell_xy near;
// 	    dist.x = poster.dist[i][j].x;
// 	    dist.y = poster.dist[i][j].y;
// 	    near.x = poster.near[i][j].x;
// 	    near.y = poster.near[i][j].y;
// 	    distmap->setDist(i, j, dist, near);
// 	  }
      }
  
  return distmap;
}

/*
 *  \brief update the local Dist Map of the Mana robot
 *  \return the resulting Dist Map if the process succeed, NULL else
 */
DistObsMap* updateDistMap(const POSTER_ID& distmapPosterId)
{
  // read the current Dist Map
  DIST_OBS_MAP distmap;

  if(MPC_avoidObsDIST_OBS_MAPPosterRead (distmapPosterId, &distmap) == ERROR)
    {
      return NULL;
    }

  // convert it into a mpc::DistObsMap class
  DistObsMap* distmapRes = DIST_OBS_MAP_POSTER_to_DISTMAP(distmap);

  return distmapRes;
}

/*
 * fill the poster containing the best control trajectory found
 */
void fillBestTrajPoster(const ManaState* s0, const ManaPCTraj* bestCTraj, const BoolObsMap* map)
{
  if(s0 != NULL && bestCTraj != NULL)
    {
      ManaPSTraj bestSTraj(*s0, *bestCTraj);  
//       ManaSTraj* sampledBestTraj = bestSTraj.sample(PL_PLAN_MAXSIZE);
//       ManaSTraj* sampledBestTraj = bestSTraj.sample(CAND_PATH_MAXSIZE);
      ManaSTraj* sampledBestTraj = bestSTraj.sample(GEN_PLAN_MAXSIZE);

      // check if the trajectory is collision-free if a map is available
      vector<bool> vec_col;
      if(map != NULL) CollisionCheck::checkCol(*map, *sampledBestTraj, vec_col);
      
      int P = sampledBestTraj->size();
      for(int k=0; k<P; k++)
	{
	  SDI_F->bestPath.tab[k].x = sampledBestTraj->at(k).x();
	  SDI_F->bestPath.tab[k].y = sampledBestTraj->at(k).y();
	  SDI_F->bestPath.tab[k].theta = sampledBestTraj->at(k).th();

	  if(map == NULL) SDI_F->bestPath.status[k] = CFG_FREE;
	  else 
	    {
	      if(vec_col[k] == true) { SDI_F->bestPath.status[k] = CFG_COL; }
	      else { SDI_F->bestPath.status[k] = CFG_FREE;  }
	    }
	}
      SDI_F->bestPath.nbPos = sampledBestTraj->size();

//       cout << "SDI_F->bestPath.nbPos : " << SDI_F->bestPath.nbPos << endl;
      
      delete sampledBestTraj;
    }
  else SDI_F->bestPath.nbPos = 0;
}

/*
 *  \brief save the ATF registered trajectories into an output file
 *  \param vec_pstraj : the input set of trajectories
 *  \param filename : the output filename
 */
void saveATF_Trajs_in_file(const vector<MANA_PSTRAJ>& vec_pstraj, char* filename)
{
  // create an output pile
  ofstream file;

  // open the output file
  file.open(filename);

  int N = vec_pstraj.size();
  file << N << endl;

  for(int i=0; i<N; i++)
    {
      /// put the current trajectory into the output file

      // get back the current genom mana pstraj
      MANA_PSTRAJ curPSTraj = vec_pstraj.at(i);

      int nbVelItv = curPSTraj.pctraj.vel.n;
      int nbAVelItv = curPSTraj.pctraj.avel.n;

      // put the current init state into the output file
      file << curPSTraj.p0.x << " ";
      file << curPSTraj.p0.y << " ";
      file << curPSTraj.p0.theta << " ";

//       if(nbVelItv == 0 || nbAVelItv == 0)
// 	{
// 	  file << 0 << " ";
// 	  file << 0;
// 	}
//       else
// 	{
// 	  file << curPSTraj.pctraj.vel.tabItv[0].poly.a << " ";
// 	  file << curPSTraj.pctraj.avel.tabItv[0].poly.a;
// 	}
//       file << endl;

      file << curPSTraj.pctraj.vel.n << " ";
      for(int i=0; i<curPSTraj.pctraj.vel.n; i++)
	{
	  file << curPSTraj.pctraj.vel.tabItv[i].itv.binf << " ";
	  file << curPSTraj.pctraj.vel.tabItv[i].itv.bsup << " ";
	  file << curPSTraj.pctraj.vel.tabItv[i].poly.a << " ";
	  file << curPSTraj.pctraj.vel.tabItv[i].poly.b << " ";
	  file << curPSTraj.pctraj.vel.tabItv[i].poly.c << " ";
	  file << curPSTraj.pctraj.vel.tabItv[i].poly.d << " ";
	}
      file << endl;

      file << curPSTraj.pctraj.avel.n << " ";
      for(int i=0; i<curPSTraj.pctraj.avel.n; i++)
	{
	  file << curPSTraj.pctraj.avel.tabItv[i].itv.binf << " ";
	  file << curPSTraj.pctraj.avel.tabItv[i].itv.bsup << " ";
	  file << curPSTraj.pctraj.avel.tabItv[i].poly.a << " ";
	  file << curPSTraj.pctraj.avel.tabItv[i].poly.b << " ";
	  file << curPSTraj.pctraj.avel.tabItv[i].poly.c << " ";
	  file << curPSTraj.pctraj.avel.tabItv[i].poly.d << " ";	
	}
      file << endl;
    }

  // close the file
  file.close();
}

/*
 *  \brief concert a couple (DD2State, DD2PCTraj) into a MANA_PSTRAJ structure
 *  \param s0 : input mpc initial state
 *  \param pctraj : input mpc control trajectory
 *  \param mana_pstraj : the genom state trajectory structure
 */
void convert_mpcStatePCTraj_into_MANA_PSTRAJ(const DD2State& s0, const DD2PCTraj& pctraj, struct timeval date, MANA_PSTRAJ& mana_pstraj)
{
  mana_pstraj.p0.x = s0.x();
  mana_pstraj.p0.y = s0.y();
  mana_pstraj.p0.theta = s0.th();

  vector<vector<double> > coefPolyVel;
  vector<vector<double> > coefPolyAVel;
  double binf; 
  double bsup;
	
  pctraj.getPolyCoef(coefPolyVel, coefPolyAVel);

  double dt = pctraj.endTime() - pctraj.startTime();
  mana_pstraj.pctraj.t0.sec = date.tv_sec;
  mana_pstraj.pctraj.t0.usec = date.tv_usec;
  mana_pstraj.pctraj.tf.sec = date.tv_sec + ((int)dt);
  mana_pstraj.pctraj.tf.usec = date.tv_usec + (dt - ((int)dt))*1000000.0;

  int VEL_N = coefPolyVel.size();
  if(VEL_N > MAX_NB_ITV)
    {
      cerr << "fillMANA_PSTRAJ_Poster : The number of intervals of the linear velocity profile is too great" << endl;
      VEL_N = MAX_NB_ITV;
//       return;
    }

  for(int i=0; i<VEL_N; i++)
    {
      int curSize = coefPolyVel.at(i).size();

      mana_pstraj.pctraj.vel.tabItv[i].itv.binf = coefPolyVel.at(i).at(0);
      mana_pstraj.pctraj.vel.tabItv[i].itv.bsup = coefPolyVel.at(i).at(1);

      if(curSize >= 3) mana_pstraj.pctraj.vel.tabItv[i].poly.a = coefPolyVel.at(i).at(2);
      else mana_pstraj.pctraj.vel.tabItv[i].poly.a = 0;
      if(curSize >= 4) mana_pstraj.pctraj.vel.tabItv[i].poly.b = coefPolyVel.at(i).at(3);
      else mana_pstraj.pctraj.vel.tabItv[i].poly.b = 0;
      if(curSize >= 5) mana_pstraj.pctraj.vel.tabItv[i].poly.c = coefPolyVel.at(i).at(4);
      else mana_pstraj.pctraj.vel.tabItv[i].poly.c = 0;
      if(curSize >= 6) mana_pstraj.pctraj.vel.tabItv[i].poly.d = coefPolyVel.at(i).at(5);
      else mana_pstraj.pctraj.vel.tabItv[i].poly.d = 0;
    }
  mana_pstraj.pctraj.vel.n = VEL_N;

  int AVEL_N = coefPolyAVel.size();
  if(AVEL_N > MAX_NB_ITV)
    {
      cerr << "fillMANA_PSTRAJ_Poster : The number of intervals of the angular velocity profile is too great" << endl;
      AVEL_N = MAX_NB_ITV;
//       return;
    }

  for(int i=0; i<AVEL_N; i++)
    {
      int curSize = coefPolyAVel.at(i).size();

      mana_pstraj.pctraj.avel.tabItv[i].itv.binf = coefPolyAVel.at(i).at(0);
      mana_pstraj.pctraj.avel.tabItv[i].itv.bsup = coefPolyAVel.at(i).at(1);

      if(curSize >= 3) mana_pstraj.pctraj.avel.tabItv[i].poly.a = coefPolyAVel.at(i).at(2);
      else mana_pstraj.pctraj.avel.tabItv[i].poly.a = 0;
      if(curSize >= 4) mana_pstraj.pctraj.avel.tabItv[i].poly.b = coefPolyAVel.at(i).at(3);
      else mana_pstraj.pctraj.avel.tabItv[i].poly.b = 0;
      if(curSize >= 5) mana_pstraj.pctraj.avel.tabItv[i].poly.c = coefPolyAVel.at(i).at(4);
      else mana_pstraj.pctraj.avel.tabItv[i].poly.c = 0;
      if(curSize >= 6) mana_pstraj.pctraj.avel.tabItv[i].poly.d = coefPolyAVel.at(i).at(5);
      else mana_pstraj.pctraj.avel.tabItv[i].poly.d = 0;
    }
  mana_pstraj.pctraj.avel.n = AVEL_N;
}
