#ifndef OA_POSTER_TTT_H
#define OA_POSTER_TTT_H

#include "server/MPC_avoidObsHeader.h"

#include <MPC/common.hpp>
#include <MPC/system.hpp>
#include <MPC/trajSel.hpp>
#include <MPC/controller.hpp>
#include <MPC/trajFollow.hpp>
#include <MPC/avoidObs.hpp>

using namespace mpc;


/* /*------------------------------------------------------------------------ */
/*  *  Compute the lateral shifts */
/*  */ 
/* void fillLateralShift(const ManaState& s0, const ManaPCTraj* pctraj, struct timeval dateBegTraj); */

/**
 *  \brief Convert a PL_PATH_XY poster into a DPMPath
 *  \param poster : the input poster
 *  \return a new DPMPath corresponding to the input poster
 *
 *  A new instance of DPMPath object is created
 */
DPMPath* PL_PATH_XYstruct_to_DPMPath(const PL_PATH_XY& poster);

/**
 *  \brief Convert a DTM_P3D poster into a DEM
 *  \param poster : the input poster
 *  \return a new DEM corresponding to the input poster
 *
 *  A new instance of DEM object is created
 */
DEM* DTM_P3D_POSTER_to_DEM(const DTM_P3D_POSTER& poster);

/*
 *  \brief update the local DEM of the Mana robot
 *  \return the resulting DEM if the process succeed, NULL else
 */
DEM* updateDEM(const POSTER_ID& demPosterId);

/*
 *  \brief Convert a DIST_OBS_MAP genom structure into a mpc DistObsMap class
 *  \param poster : the input genom DIST_OBS_MAP poster
 *  \return a mpc distance to obstacles map
 */
DistObsMap* DIST_OBS_MAP_POSTER_to_DISTMAP(const DIST_OBS_MAP& poster);

/*
 *  \brief update the local Dist Map of the Mana robot
 *  \return the resulting Dist Map if the process succeed, NULL else
 */
DistObsMap* updateDistMap(const POSTER_ID& distmapPosterId);

/* /* */
/*  *  \brief update the localization of the Mana robot */
/*  *  \param the resulting state */
/*  *  \return 0 if the process succeed, -1 else */
/*  */ 
/* int updateManaLoc(ManaState &sres); */

/*
 * \brief fill posters displaying the candidate trajectory and the best one found
 */
void fillCandTrajPosters(ManaTrajSel* trajSel);

/*
 * \brief fill the poster containing the best control trajectory found
 */ 
void fillBestTrajPoster(const ManaState* s0, const ManaPCTraj* bestCTraj, const BoolObsMap* map = NULL);

/* /* */
/*  * fill the poster containing the best control trajectory found */
/*  */ 
/* void fillBestCTrajPoster(ManaPCTraj* bestCTraj); */

/*
 *  \brief fill the poster containing the waypoints computed by the mpc avoid traj following process
 *  \param mpc_waypoints : mpc waypoints
 */ 
void fillATFwaypoints(const std::vector<WayPoint>& mpc_waypoints);

/*
 *  \brief Fill the distance to obstacles map poster from a mpc distmap
 *  \param distMap : the input mpc distmap
 */
int fillDistMapPoster(DistObsMap* distMap);

/*
 *  \brief save the ATF registered trajectories into an output file
 *  \param vec_pstraj : the input set of trajectories
 *  \param filename : the output filename
 */
void saveATF_Trajs_in_file(const vector<MANA_PSTRAJ>& vec_pstraj, char* filename);

/*
 *  \brief convert a couple (DD2State, DD2PCTraj) into a MANA_PSTRAJ structure
 *  \param s0 : input mpc initial state
 *  \param pctraj : input mpc control trajectory
 *  \param mana_pstraj : the genom state trajectory structure
 */
void convert_mpcStatePCTraj_into_MANA_PSTRAJ(const DD2State& s0, const DD2PCTraj& pctraj, struct timeval date, MANA_PSTRAJ& mana_pstraj);


#endif // OA_POSTER_TTT_H
